@@ -23,8 +23,8 @@ class WhyNot(Pass):
         self.why_meta = WhyMeta(task)
         self.walk(task)
 
-    def query(self, task: Task, parent=None):
-        for item in task.items:
+    def query(self, query: Task, parent=None):
+        for item in query.items:
             if item.action == ActionType.Get:
                 self.explode_get(item, parent)
 
