@@ -72,9 +72,12 @@ class Resources(ResourcesBase):
     #--------------------------------------------------
     # Engines
     #--------------------------------------------------
-    def get_engine_sizes(self):
+    def get_engine_sizes(self, cloud_provider: str|None=None):
         return ENGINE_SIZES
 
+    def get_cloud_provider(self) -> str:
+        return "azure"
+
     def list_engines(self, state:str|None = None):
         return api.list_engines(self._api_ctx(), state)
 
