@@ -159,7 +159,7 @@ class ResourcesBase(ABC):
     #--------------------------------------------------
 
     @abstractmethod
-    def get_engine_sizes(self) -> List[Any]:
+    def get_engine_sizes(self, cloud_provider: str|None=None) -> List[Any]:
         pass
 
     @abstractmethod
@@ -170,6 +170,10 @@ class ResourcesBase(ABC):
     def get_engine(self, name: str) -> EngineState | None:
         pass
 
+    @abstractmethod
+    def get_cloud_provider(self) -> str:
+        pass
+
     @abstractmethod
     def is_valid_engine_state(self, name: str) -> bool:
         pass
