@@ -21,3 +21,12 @@ def sqrt(value: _Number) -> b.Expression:
 
 def maximum(left: _Number, right: _Number) -> b.Expression:
     return _make_expr("maximum", left, right, b.Number.ref("res"))
+
+def minimum(left: _Number, right: _Number) -> b.Expression:
+    return _make_expr("minimum", left, right, b.Number.ref("res"))
+
+def isinf(value: _Number) -> b.Expression:
+    return _make_expr("isinf", value)
+
+def isnan(value: _Number) -> b.Expression:
+    return _make_expr("isnan", value)
