@@ -1,12 +1,13 @@
+import warnings
 from collections import defaultdict
 from relationalai.early_access.dsl.adapters.orm.model import ExclusiveInclusiveSubtypeFact, ExclusiveSubtypeFact, \
     ORMRingType, ORMValueComparisonOperator
 from relationalai.early_access.dsl.adapters.orm.parser import ORMParser
 from relationalai.early_access.dsl.core.utils import to_pascal_case
-from relationalai.early_access.dsl.orm.constraints import RingType, ValueComparisonType
+from relationalai.early_access.dsl.orm.constraints import RingType, ValueComparisonType, Range
 from relationalai.early_access.dsl.orm.models import Model
 from relationalai.early_access.builder import Integer, String, DateTime, Decimal, Date, Float
-from relationalai.early_access.dsl.orm.utils import build_relation_name_from_reading
+from relationalai.early_access.metamodel.util import NameCache
 
 
 class ORMAdapterQB:
@@ -58,6 +59,7 @@ class ORMAdapterQB:
         self._parser = ORMParser(orm_file_path)
         self._relationship_role_value_constraints = defaultdict()
         self._relationships = {}
+        self.name_cache = NameCache()
         self.model = self.orm_to_model()
 
     def orm_to_model(self):
@@ -84,18 +86,23 @@ class ORMAdapterQB:
             if len(rvc.values) > 0 and isinstance(rvc.values[0], str):
                 enum_to_values[rvc.role.player].extend(rvc.values)
         # create an Enum instead of ValueType when there is a role value constraint (only string sequence) played by this ValueType
+        cardinality_constraints = self._parser.objects_type_to_cardinality_constraints()
         for vt in self._parser.value_types().values():
             if vt.id in enum_to_values:
-                model.Enum(to_pascal_case(vt.name), enum_to_values[vt.id])
+                concept = model.Enum(to_pascal_case(vt.name), enum_to_values[vt.id])
             else:
-                model.Concept(vt.name, extends=[self._datatype_mapping.get(vt.data_type, String)])
+                concept = model.Concept(vt.name, extends=[self._datatype_mapping.get(vt.data_type, String)])
+            if vt.name in cardinality_constraints:
+                model.cardinality(concept, self._build_constraint_ranges(cardinality_constraints[vt.name]))
 
     def _add_entity_types(self, model):
         extended_concepts = [y.subtype_name for x in self._parser.subtype_facts().values() for y in x]
-
+        cardinality_constraints = self._parser.objects_type_to_cardinality_constraints()
         for et in self._parser.entity_types().values():
             if et.name not in extended_concepts:
-                model.Concept(et.name)
+                concept = model.Concept(et.name)
+                if et.name in cardinality_constraints:
+                    model.cardinality(concept, self._build_constraint_ranges(cardinality_constraints[et.name]))
 
         subtype_facts = self._parser.sorted_subtype_facts()
         for st_fact in subtype_facts:
@@ -103,7 +110,12 @@ class ORMAdapterQB:
             parent_entity = model.lookup_concept(parent)
             if parent_entity is None:
                 parent_entity = model.Concept(parent)
-            model.Concept(st_fact.subtype_name, extends=[parent_entity])
+                if parent in cardinality_constraints:
+                    model.cardinality(parent_entity, self._build_constraint_ranges(cardinality_constraints[parent]))
+            child = st_fact.subtype_name
+            child_entity = model.Concept(child, extends=[parent_entity])
+            if child in cardinality_constraints:
+                model.cardinality(child_entity, self._build_constraint_ranges(cardinality_constraints[child]))
 
     def _add_subtype_relationships(self, model):
         for parent, children in self._parser.subtype_facts().items():
@@ -136,11 +148,16 @@ class ORMAdapterQB:
             rdo = reading_orders[0]
             player = object_types[rdo.roles[0].role.player].name
             player_entity = model.lookup_concept(player)
-            rel_name, reading = self._build_relation(model, fact_type, player_entity, rdo)
+            reading = self._build_reading(model, rdo)
             relationship = model.Relationship(reading)
-            setattr(player_entity, rel_name, relationship)
+            self._set_role_verbalization(relationship, rdo)
+            short_name = self._pick_short_name(model, player_entity, fact_type, relationship._readings[0])
+            cached_name = self.name_cache.get_name(relationship._readings[0], f"{player_entity._name}.{short_name}")
+            short_name = cached_name.split(".")[1]
+            setattr(player_entity, short_name, relationship)
             self._relationships[fact_type] = relationship
 
+
             # Marking unique and mandatory roles
             role_idx_to_player = []
             for ro in rdo.roles:
@@ -156,12 +173,18 @@ class ORMAdapterQB:
                     self._relationship_role_value_constraints[role] = role_value_constraints[role_id].values
 
             # Adding alternative readings
+            self._validate_role_verbalization(reading_orders)
+
             if len(reading_orders) > 1:
                 for rdo in reading_orders[1:]:
                     other_player = object_types[rdo.roles[0].role.player].name
                     other_player_entity = model.lookup_concept(other_player)
-                    rel_name, alt_reading = self._build_relation(model, fact_type, other_player_entity, rdo)
-                    setattr(other_player_entity, rel_name, relationship.alt(alt_reading))
+                    alt_reading = self._build_reading(model, rdo)
+                    alt_reading_obj = relationship.alt(alt_reading)
+                    short_name = self._pick_short_name(model, other_player_entity, fact_type, alt_reading_obj)
+                    cached_name = self.name_cache.get_name(alt_reading_obj, f"{other_player_entity._name}.{short_name}")
+                    short_name = cached_name.split(".")[1]
+                    setattr(other_player_entity, short_name, alt_reading_obj)
 
             # Marking identifying relationships
             if fact_type in fact_type_to_internal_ucs:
@@ -183,20 +206,27 @@ class ORMAdapterQB:
 
     def _add_external_identifying_relationships(self, model):
         roles = self._parser.roles()
-        fact_types_to_readings = self._parser.fact_type_readings()
         object_types = self._parser.object_types()
-        for uc_id, uc in self._parser.external_uniqueness_constraints().items():
-            entity = model.lookup_concept(object_types[uc.identifies].name)
-            identifying_rel = []
-            for ro in uc.roles:
-                fact_name = roles[ro].relationship_name
-                readings = fact_types_to_readings[fact_name]
-                for rdo in readings:
-                    if rdo.roles[0].role.player == uc.identifies:
-                        for rl in rdo.roles:
-                            if rl.role.id == ro:
-                                identifying_rel.append(self._relationships[fact_name])
-            entity.identify_by(*identifying_rel)
+        for uc in self._parser.external_uniqueness_constraints().values():
+            # Identifying external UCs
+            if uc.identifies:
+                entity = model.lookup_concept(object_types[uc.identifies].name)
+                identifying_rel = []
+                for ro in uc.roles:
+                    relationship = self._relationships[roles[ro].relationship_name]
+                    # TODO: this will change when we can use readings or roles with identify_by
+                    if relationship._roles()[0]._concept._name != entity._name:
+                        raise ValueError(f"The concept {entity._name} must be the first player in all its identifying relationships.")
+                    identifying_rel.append(relationship)
+                entity.identify_by(*identifying_rel)
+            # Non identifying external UCs
+            else:
+                role_list = []
+                for ro in uc.roles:
+                    relationship = self._relationships[roles[ro].relationship_name]
+                    role = relationship._rel_roles[object_types[roles[ro].player].name.lower()]
+                    role_list.append(role)
+                model.unique(*role_list)
 
     def _add_inclusive_role_constraints(self, model):
         for rc in self._parser.inclusive_role_constraints():
@@ -243,15 +273,13 @@ class ORMAdapterQB:
     def _build_reading(self, model, reading_order):
         object_types = self._parser.object_types()
         rel_args = []
+        if reading_order.front_text is not None:
+            rel_args.append(f"{reading_order.front_text} ")
         for rdo_role in reading_order.roles:
-            if rdo_role.prefix is not None:
-                rel_args.append(rdo_role.prefix + "-")
             p = model.lookup_concept(object_types[rdo_role.role.player].name)
             rel_args.append(f"{{{p}}}") if rdo_role.role.name == "" else rel_args.append(f"{{{rdo_role.role.name}:{p}}}")
-            if rdo_role.postfix is not None:
-                rel_args.append("-" + rdo_role.postfix)
             if rdo_role.text is not None:
-                rel_args.append(" " + rdo_role.text + " ")
+                rel_args.append(f" {rdo_role.text} ")
         return ''.join(rel_args).strip()
 
     def _rename_to_ref_mode(self, model, first_player_entity, fact_type):
@@ -263,19 +291,13 @@ class ORMAdapterQB:
                 return entity_type.ref_mode
         return None
 
-    def _build_relation(self, model, fact_type, player_entity, reading_order):
-        object_types = self._parser.object_types()
-        role_name = ""
-        for role in reading_order.roles:
-            if object_types[role.role.player].name != player_entity._name:
-                role_name = role.role.name
-        reading = self._build_reading(model, reading_order)
+    def _pick_short_name(self, model, player_entity, fact_type, reading):
         ref_mode = self._rename_to_ref_mode(model, player_entity, fact_type)
         if ref_mode is not None:
             rel_name = ref_mode.lower()
         else:
-            rel_name = role_name if role_name != "" else build_relation_name_from_reading(reading)
-        return rel_name, reading
+            rel_name = reading.rai_way_name()
+        return rel_name
 
     def _role_id_to_role_object(self, model, role_obj):
         player_entity = model.lookup_concept(self._parser.object_types()[role_obj.player].name)
@@ -294,3 +316,40 @@ class ORMAdapterQB:
         # Role list
         else:
             return [self._role_id_to_role_object(model, roles[ro]) for ro in orm_constraint.roles]
+
+    @staticmethod
+    def _build_constraint_ranges(constraint):
+        constraint_ranges = []
+        for rg in constraint.ranges:
+            if rg.range_from == rg.range_to:
+                constraint_ranges.append(int(rg.range_from))
+            else:
+                constraint_ranges.append(Range.between(int(rg.range_from), int(rg.range_to)))
+        return constraint_ranges
+
+    @staticmethod
+    def _set_role_verbalization(relationship, rdo):
+        for idx, ro in enumerate(rdo.roles):
+            qb_role = relationship._roles()[idx]
+            text_chunks = [None, None]
+            if ro.prefix:
+                text_chunks[0] = ro.prefix
+            if ro.postfix:
+                text_chunks[1] = ro.postfix
+            if text_chunks[0] or text_chunks[1]:
+                qb_role.verbalization(*text_chunks)
+
+    @staticmethod
+    def _validate_role_verbalization(reading_orders):
+        if len(reading_orders) > 1:
+            relationship = reading_orders[0].roles[0].role.relationship_name
+            for ro in reading_orders[0].roles:
+                prefix = ro.prefix
+                postfix = ro.postfix
+                player = ro.role.player
+                #  A role player must have the same prefix in all readings
+                if not all(alt_ro.prefix == prefix for rdo in reading_orders[1:] for alt_ro in rdo.roles if alt_ro.role.player == player):
+                    warnings.warn(f"Ignoring incompatible prefix in alternative readings of {relationship}.")
+                #  A role player must have the same postfix in all readings
+                if not all(alt_ro.postfix == postfix for rdo in reading_orders[1:] for alt_ro in rdo.roles if alt_ro.role.player == player):
+                    warnings.warn(f"Ignoring incompatible postfix in alternative readings of {relationship}")
