@@ -4,11 +4,12 @@ import re
 import xmltodict
 from collections import defaultdict
 from relationalai.early_access.dsl.adapters.orm.model import ORMEntityType, ORMValueType, ORMRole, ORMSubtypeFact, \
-    ORMUniquenessConstraint, ORMExclusionConstraint, ORMMandatoryConstraint, ORMReadingOrderRole, ORMReadingOrder, \
+    ORMUniquenessConstraint, ORMExclusionConstraint, ORMMandatoryConstraint, ORMReadingRole, ORMReading, \
     ExclusiveInclusiveSubtypeFact, SubtypeFact, ExclusiveSubtypeFact, ORMInclusionConstraint, \
     ORMRoleValueConstraint, ORMValueComparisonConstraint, ORMValueComparisonOperator, ORMRingConstraint, ORMRingType, \
     ORMRoleSubsetConstraint, ORMInclusiveRoleConstraint, ORMExclusiveRoleConstraint, ORMEqualityConstraint, \
-    ORMFrequencyConstraint
+    ORMFrequencyConstraint, ORMCardinalityConstraint, ORMRange
+from relationalai.early_access.metamodel.util import ordered_set
 from relationalai.util.graph import topological_sort
 
 
@@ -20,6 +21,9 @@ class ORMParser:
 
         self._ontology = json.loads(json.dumps(data))
 
+        self._role_to_player = dict()
+        self._cardinality_constraints = dict()
+
         self._model_name = self._parse_model_name()
         self._entity_types = self._parse_entity_types()
         self._value_types = self._parse_value_types()
@@ -44,6 +48,7 @@ class ORMParser:
         self._sorted_subtype_facts = self._sort_subtype_facts()
         self._fact_type_readings = self._parse_fact_types_reading_orders()
         self._mandatory_roles = self._parse_mandatory_roles()
+        self._object_types_to_cardinality_constraints = self._parse_object_types_to_cardinality_constraints()
 
     def model_name(self):
         return self._model_name
@@ -96,6 +101,9 @@ class ORMParser:
     def equality_constraints(self):
         return self._equality_constraints
 
+    def cardinality_constraints(self):
+        return self._cardinality_constraints
+
     def unique_roles(self):
         return self._unique_roles
 
@@ -120,6 +128,9 @@ class ORMParser:
     def sorted_subtype_facts(self):
         return self._sorted_subtype_facts
 
+    def objects_type_to_cardinality_constraints(self):
+        return self._object_types_to_cardinality_constraints
+
     def _parse_fact_type_to_roles(self):
         fact_type_data = defaultdict(list)
         for role in self._roles.values():
@@ -127,6 +138,13 @@ class ORMParser:
             fact_type_data[relationship_name].append(role)
         return fact_type_data
 
+    def _parse_object_types_to_cardinality_constraints(self):
+        object_types_constraints = {}
+        for cc in self._cardinality_constraints.values():
+            object_type_name = self._object_types[cc.object_type].name
+            object_types_constraints[object_type_name] = cc
+        return object_types_constraints
+
     def _parse_model_name(self):
         model_name = self._get_nested(self._ontology, "ormRoot:ORM2", "orm:ORMModel", "@Name")
         return model_name if model_name else "ORMModel"
@@ -141,6 +159,8 @@ class ORMParser:
                 name = et['@Name']
                 ref_mode = et.get("@_ReferenceMode") or None
                 entity_types[id] = ORMEntityType(id, name, ref_mode)
+                self._parse_role_to_player_reference(et)
+                self._parse_cardinality_constraint(et)
         return entity_types
 
     def _parse_value_types(self):
@@ -155,6 +175,8 @@ class ORMParser:
                     name = vt["@Name"]
                     data_type = data_types[self._get_nested(vt, "orm:ConceptualDataType", "@ref")]
                     value_types[id] = ORMValueType(id, name, data_type)
+                    self._parse_role_to_player_reference(vt)
+                    self._parse_cardinality_constraint(vt)
         return value_types
 
     def _parse_data_types(self):
@@ -168,7 +190,6 @@ class ORMParser:
 
     def _parse_roles(self):
         roles = {}
-        role_to_player = self._parse_role_player()
         orm_fact_types = self._get_nested(self._ontology, "ormRoot:ORM2", "orm:ORMModel", "orm:Facts", "orm:Fact")
         if orm_fact_types:
             for ft in self._single_object_to_list(orm_fact_types):
@@ -180,24 +201,14 @@ class ORMParser:
                         role_name = ro["@Name"]
                         role_player = self._get_nested(ro, "orm:RolePlayer", "@ref")
                         if role_player in self._object_types:
-                            roles[role_id] = ORMRole(role_id, role_name, relationship_name, role_to_player[role_id])
+                            roles[role_id] = ORMRole(role_id, role_name, relationship_name, self._role_to_player[role_id])
         return roles
 
-    def _parse_role_player(self):
-        role_to_player = {}
-        orm_entity_types = (self._get_nested(self._ontology, "ormRoot:ORM2", "orm:ORMModel", "orm:Objects",
-                                          "orm:EntityType") or [])
-        orm_value_types = (self._get_nested(self._ontology, "ormRoot:ORM2", "orm:ORMModel", "orm:Objects",
-                         "orm:ValueType") or [])
-        object_types = self._single_object_to_list(orm_entity_types) + self._single_object_to_list(orm_value_types)
-
-        if object_types:
-            for ot in object_types:
-                roles = self._get_nested(ot, "orm:PlayedRoles", "orm:Role")
-                if roles:
-                    for role in self._single_object_to_list(roles):
-                        role_to_player[role["@ref"]] = ot["@id"]
-        return role_to_player
+    def _parse_role_to_player_reference(self, object_type):
+        roles = self._get_nested(object_type, "orm:PlayedRoles", "orm:Role")
+        if roles:
+            for role in self._single_object_to_list(roles):
+                self._role_to_player[role["@ref"]] = object_type["@id"]
 
     def _parse_role_value_constraints(self):
         role_value_constraints = {}
@@ -409,43 +420,62 @@ class ORMParser:
                             fact_types_readings[relationship_name].append(reading_order)
         return fact_types_readings
 
-
     def _parse_fact_types_reading_order(self, reading_order):
-        role_texts = self._get_nested(reading_order, "orm:Readings", "orm:Reading", "orm:ExpandedData", "orm:RoleText")
-        if not role_texts:
-            return None
+        expanded_data = self._get_nested(reading_order, "orm:Readings", "orm:Reading", "orm:ExpandedData")
+        if expanded_data:
+            front_text = expanded_data.get("@FrontText", None)
+            if front_text:
+                front_text = front_text.strip()
 
-        # Step 1: Build roles with whatever RoleText exists
-        index_to_role = {}
-        for rt in self._single_object_to_list(role_texts):
-            index = int(rt["@RoleIndex"])
-            index_to_role[index] = ORMReadingOrderRole(
-                index=index,
-                prefix=rt.get("@PreBoundText", "").strip() or None,
-                postfix=rt.get("@PostBoundText", "").strip() or None,
-                text=rt.get("@FollowingText", "").strip() or None,
-                role=None
-            )
+            role_texts = self._get_nested(expanded_data,"orm:RoleText")
+            if not role_texts:
+                return None
 
-        # Step 2: Ensure all roles from RoleSequence are represented, even if no RoleText exists
-        roles = self._get_nested(reading_order, "orm:RoleSequence", "orm:Role")
-        if roles:
-            for i, role in enumerate(self._single_object_to_list(roles)):
-                role_id = role.get("@ref")
-                orm_role = self._roles.get(role_id, None)
-                if i in index_to_role:
-                    index_to_role[i].role = orm_role
-                else:
-                    # Fill in with None/defaults if RoleText is missing
-                    index_to_role[i] = ORMReadingOrderRole(
-                        index=i,
-                        prefix=None,
-                        postfix=None,
-                        text=None,
-                        role=orm_role
-                    )
+            # Step 1: Build roles with whatever RoleText exists
+            index_to_role = {}
+            for rt in self._single_object_to_list(role_texts):
+                index = int(rt["@RoleIndex"])
+                index_to_role[index] = ORMReadingRole(
+                    index=index,
+                    prefix=rt.get("@PreBoundText", "").strip() or None,
+                    postfix=rt.get("@PostBoundText", "").strip() or None,
+                    text=rt.get("@FollowingText", "").strip() or None,
+                    role=None
+                )
 
-        return ORMReadingOrder([index_to_role[i] for i in sorted(index_to_role)])
+            # Step 2: Ensure all roles from RoleSequence are represented, even if no RoleText exists
+            roles = self._get_nested(reading_order, "orm:RoleSequence", "orm:Role")
+            if roles:
+                for i, role in enumerate(self._single_object_to_list(roles)):
+                    role_id = role.get("@ref")
+                    orm_role = self._roles.get(role_id, None)
+                    if i in index_to_role:
+                        index_to_role[i].role = orm_role
+                    else:
+                        # Fill in with None/defaults if RoleText is missing
+                        index_to_role[i] = ORMReadingRole(
+                            index=i,
+                            prefix=None,
+                            postfix=None,
+                            text=None,
+                            role=orm_role
+                        )
+
+            return ORMReading(front_text, [index_to_role[i] for i in sorted(index_to_role)])
+
+    def _parse_cardinality_constraint(self, object_type):
+        orm_cc = self._get_nested(object_type, "orm:CardinalityRestriction", "orm:CardinalityConstraint")
+        if orm_cc:
+            object_type_id = object_type.get("@id")
+            cc_id = orm_cc.get("@id")
+            ranges = self._get_nested(orm_cc, "orm:Ranges", "orm:CardinalityRange")
+            # At least a range is always present within a cardinality constraint
+            orm_ranges = []
+            for rg in self._single_object_to_list(ranges):
+                range_from = rg.get("@From")
+                range_to = rg.get("@To")
+                orm_ranges.append(ORMRange(range_from, range_to))
+            self._cardinality_constraints[cc_id] = ORMCardinalityConstraint(cc_id, object_type_id, orm_ranges)
 
     @staticmethod
     def _get_nested(d, *keys):
@@ -499,15 +529,17 @@ class ORMParser:
         # Build a dependency graph to sort subtype facts so that child entity are declared
         # after the parent entity that they extend
         subtype_facts = self._subtype_facts
-        nodes = []
         edges = []
+        nodes = ordered_set()
         sorted_subtype_facts = []
         # Parent depends on child, i.e., parent entity must be declared before child entity
         for parent in subtype_facts:
-            nodes.append(parent)
+            nodes.add(parent)
             for child in subtype_facts[parent]:
-                edges.append((parent, child.subtype_name))
-        for parent in topological_sort(nodes, edges):
+                target = child.subtype_name
+                edges.append((parent, target))
+                nodes.add(target)
+        for parent in topological_sort(list(nodes), edges):
             sorted_subtype_facts.extend(subtype_facts[parent])
         return sorted_subtype_facts
 
