@@ -1,7 +1,7 @@
 from typing import Union
 
-from relationalai.early_access.dsl.bindings.common import Binding, IdentifierConceptBinding, RoleBinding, PartitioningConceptBinding, \
-    BindableTable, BindableAttribute, ReferentConceptBinding
+from relationalai.early_access.dsl.bindings.common import Binding, IdentifierConceptBinding, RoleBinding, \
+    SubtypeConceptBinding, BindableTable, BindableAttribute, ReferentConceptBinding
 from relationalai.early_access.dsl.codegen.common import BoundIdentifierConstraint, BoundRelationship
 from relationalai.early_access.dsl.codegen.reasoner import OntologyReasoner
 from relationalai.early_access.dsl.orm.relationships import Role, Relationship
@@ -17,7 +17,7 @@ class Binder:
         # bindings classification
         self._identifier_bindings: OrderedSet[IdentifierConceptBinding] = OrderedSet()
         self._subtype_bindings: OrderedSet[Union[IdentifierConceptBinding, ReferentConceptBinding]] = OrderedSet()
-        self._filtering_subtype_bindings: OrderedSet[PartitioningConceptBinding] = OrderedSet()
+        self._subtype_concept_bindings: OrderedSet[SubtypeConceptBinding] = OrderedSet()
         self._referent_concept_bindings: OrderedSet[ReferentConceptBinding] = OrderedSet()
         self._value_type_bindings: OrderedSet[Binding] = OrderedSet()
         self._entity_type_bindings: OrderedSet[Binding] = OrderedSet()
@@ -54,21 +54,26 @@ class Binder:
     @staticmethod
     def is_filtering_binding(binding: Binding) -> bool:
         """
-        Returns True if the binding is a FilteringConceptBinding or has filter_by.
+        Returns True if the binding is a SubtypeConceptBinding or has filter_by.
         """
-        return isinstance(binding, PartitioningConceptBinding) or binding.filter_by is not None
+        return isinstance(binding, SubtypeConceptBinding) or binding.filter_by is not None
 
-    def subtype_bindings(self) -> OrderedSet[Union[IdentifierConceptBinding, ReferentConceptBinding]]:
+    def subtype_ctor_bindings(self) -> OrderedSet[Union[IdentifierConceptBinding, ReferentConceptBinding]]:
         """
-        Returns the subtype bindings.
+        Returns the identifier or referent subtype bindings. Those are only possible for subtypes that define
+        their own reference scheme.
+
+        They are different from SubtypeConceptBinding, which is a subtype binding with a filter.
         """
         return self._subtype_bindings
 
-    def filtering_subtype_bindings(self) -> OrderedSet[PartitioningConceptBinding]:
+    def subtype_filtering_bindings(self) -> OrderedSet[SubtypeConceptBinding]:
         """
-        Returns the filtering subtype bindings.
+        Returns the subtype concept bindings. These are bindings that are used to derive subtypes from a parent type,
+        and they are different from IdentifierConceptBinding or ReferentConceptBinding, which are used to construct
+        entities or refer to them.
         """
-        return self._filtering_subtype_bindings
+        return self._subtype_concept_bindings
 
     def referent_concept_bindings(self) -> OrderedSet[ReferentConceptBinding]:
         """
@@ -80,7 +85,7 @@ class Binder:
         if isinstance(binding, RoleBinding):
             # simple case: use the Role from the binding
             role = binding.role
-        elif isinstance(binding, (IdentifierConceptBinding, PartitioningConceptBinding, ReferentConceptBinding)):
+        elif isinstance(binding, (IdentifierConceptBinding, SubtypeConceptBinding, ReferentConceptBinding)):
             # either an IdentifierBinding or a SubtypeBinding, look up the role in the IdentifierRelationship
             concept = binding.entity_type
             if self._reasoner.has_own_ref_scheme(concept):
@@ -136,7 +141,7 @@ class Binder:
                 return binding
         raise ValueError(f"Cannot look up entity constructor binding for {concept} from {source}")
 
-    def lookup_subtype_reference_binding(self, binding: IdentifierConceptBinding):
+    def lookup_subtype_reference_binding(self, binding: Union[IdentifierConceptBinding, SubtypeConceptBinding]):
         """
         Look up a constructor binding referenced by the subtype binding.
         """
@@ -146,7 +151,7 @@ class Binder:
         parent_type = sibling_role.player()
         return self.lookup_constructor_binding(parent_type, binding.column)
 
-    def lookup_subtype_reference_binding_by_source(self, binding: Union[IdentifierConceptBinding, PartitioningConceptBinding]):
+    def lookup_subtype_reference_binding_by_source(self, binding: Union[IdentifierConceptBinding, SubtypeConceptBinding]):
         """
         Look up a constructor binding referenced by the subtype binding.
         """
@@ -165,10 +170,18 @@ class Binder:
         for bound_constraint in self._bound_identifier_constraints:
             if bound_constraint.constraint in identifier_constraints and bound_constraint.table == source:
                 bound_constraints.add(bound_constraint)
-        if len(bound_constraints) != len(identifier_constraints):
-            raise ValueError(f"Cannot look up bound identifier constraints for {concept} from {source}, "
-                             f"expected {len(identifier_constraints)} but found {len(bound_constraints)} "
-                             f"constraints")
+        #=
+        # Note: this is a bit naive right now, it should take into account whether the concept is being:
+        #  * constructed (we need ALL) or;
+        #  * referenced (we need exactly one).
+        #
+        # To be addressed in https://relationalai.atlassian.net/browse/RAI-39757.
+        #=
+        if not (bound_constraints and all(b.constraint in identifier_constraints for b in bound_constraints)):
+            raise ValueError(
+                f"Cannot look up bound identifier constraints for {concept} from {source}, "
+                f"no bound constraints found"
+            )
         return bound_constraints
 
     def bound_identifier_constraints(self):
@@ -219,8 +232,8 @@ class Binder:
             self._process_concept_binding(binding)
         elif isinstance(binding, RoleBinding):
             self._process_role_binding(binding)
-        elif isinstance(binding, PartitioningConceptBinding):
-            self._process_filtering_subtype_binding(binding)
+        elif isinstance(binding, SubtypeConceptBinding):
+            self._process_subtype_concept_binding(binding)
         else:
             raise ValueError(f"Unknown binding type: {type(binding)}")
 
@@ -257,19 +270,18 @@ class Binder:
         self._classify_binding_by_type(binding, binding.role.player())
         self._role_bindings.setdefault(binding.role, OrderedSet()).add(binding)
 
-    def _process_filtering_subtype_binding(self, binding: PartitioningConceptBinding):
+    def _process_subtype_concept_binding(self, binding: SubtypeConceptBinding):
         """
         FilteringSubtypeBinding is a subtype binding with a filter.
 
         It is neither an entity nor a value type binding. The filter may be a literal or a value from a
         RoleValueConstraint.
         """
-        self._filtering_subtype_bindings.add(binding)
-        # TODO: add to self._role_bindings
+        self._subtype_concept_bindings.add(binding)
 
     def _classify_binding_by_type(self, binding: Binding, player: qb.Concept):
-        if isinstance(binding, PartitioningConceptBinding):
-            # FilteringSubtypeBinding is neither an entity nor a value type binding.
+        if isinstance(binding, SubtypeConceptBinding):
+            # SubtypeConceptBinding is neither an entity nor a value type binding.
             return
         if player._is_primitive():
             self._value_type_bindings.add(binding)
@@ -384,7 +396,8 @@ class Binder:
 
         has_entity_type_player, is_bound = False, True
         for role in roles:
-            if role not in bound_roles:
+            if not (role in bound_roles or
+                    self.lookup_bound_identifier_constraint_by_source(role.player(), bound_relationship.table)):
                 is_bound = False
                 break
             if not role.player()._is_primitive():
