@@ -5,7 +5,7 @@ from typing import Union, Optional
 
 from relationalai.early_access.builder import where
 from relationalai.early_access.dsl.bindings.common import Binding, RoleBinding, IdentifierConceptBinding, BindableTable, \
-    BindableAttribute, PartitioningConceptBinding, ReferentConceptBinding
+    BindableAttribute, SubtypeConceptBinding, ReferentConceptBinding
 from relationalai.early_access.dsl.codegen.binder import Binder
 from relationalai.early_access.dsl.codegen.common import BoundIdentifierConstraint, BoundRelationship
 from relationalai.early_access.dsl.codegen.reasoner import OntologyReasoner
@@ -40,7 +40,7 @@ class Weaver:
         self._bound_identifier_to_composite_entity_map:\
             dict[BoundIdentifierConstraint, OrderedSet[CompositeEntityMap]] = OrderedDict()
         self._subtype_binding_to_entity_map: dict[Union[IdentifierConceptBinding, ReferentConceptBinding], EntitySubtypeMap] = OrderedDict()
-        self._filtering_subtype_binding_to_entity_map: dict[PartitioningConceptBinding, EntitySubtypeMap] = OrderedDict()
+        self._filtering_subtype_binding_to_entity_map: dict[SubtypeConceptBinding, EntitySubtypeMap] = OrderedDict()
         self._supertype_to_entity_map: dict[qb.Concept, UnionEntityMap] = OrderedDict()
 
     def generate(self):
@@ -255,7 +255,7 @@ class Weaver:
         return matching_entity_maps
 
     def _generate_subtype_entity_maps(self):
-        for binding in self._binder.subtype_bindings():
+        for binding in self._binder.subtype_ctor_bindings():
             self._generate_subtype_entity_map(binding)
 
     def _generate_subtype_entity_map(self, binding: Union[IdentifierConceptBinding, ReferentConceptBinding]):
@@ -268,12 +268,12 @@ class Weaver:
         self._subtype_binding_to_entity_map[binding] = entity_map
 
     def _generate_filtered_subtype_entity_maps(self):
-        for binding in self._binder.filtering_subtype_bindings():
+        for binding in self._binder.subtype_filtering_bindings():
             self._generate_filtered_subtype_entity_map(binding)
 
-    def _generate_filtered_subtype_entity_map(self, binding: PartitioningConceptBinding):
-        ctor_binding = self._binder.lookup_subtype_reference_binding_by_source(binding)
-        ctor_entity_map = self._ctor_binding_to_entity_map[ctor_binding]
+    def _generate_filtered_subtype_entity_map(self, binding: SubtypeConceptBinding):
+        ctor_binding = self._binder.lookup_subtype_reference_binding(binding)
+        ctor_entity_map = self._lookup_ref_role_map_by_ctor_binding(ctor_binding)
         entity_map = EntitySubtypeMap(self._model, binding, ctor_entity_map)
         self._filtering_subtype_binding_to_entity_map[binding] = entity_map
 
