@@ -3,7 +3,8 @@ import typing
 from relationalai.util.graph import topological_sort
 from relationalai.early_access.dsl.orm.constraints import Unique, Mandatory, RoleValueConstraint, \
     InclusiveSubtypeConstraint, ExclusiveSubtypeConstraint, Range, RingConstraint, InclusiveRoleConstraint, \
-    ExclusiveRoleConstraint, ValueComparisonConstraint, RoleSubsetConstraint, EqualityConstraint, FrequencyConstraint
+    ExclusiveRoleConstraint, ValueComparisonConstraint, RoleSubsetConstraint, EqualityConstraint, FrequencyConstraint, \
+    CardinalityConstraint
 from relationalai.early_access.dsl.orm.models import Model
 from relationalai.early_access.dsl.orm.relationships import Role
 from relationalai.early_access.dsl.orm.types import Concept
@@ -25,6 +26,7 @@ class Printer(BasePrinter):
     _role_subset_constraints: OrderedSet[RoleSubsetConstraint]
     _equality_constraints: OrderedSet[EqualityConstraint]
     _frequency_constraints: OrderedSet[FrequencyConstraint]
+    _cardinality_constraints: OrderedSet[CardinalityConstraint]
 
     def __init__(self, io: typing.Optional[typing.IO[str]] = None):
         # Set the base class field (frozen)
@@ -43,6 +45,7 @@ class Printer(BasePrinter):
         object.__setattr__(self, '_role_subset_constraints', OrderedSet())
         object.__setattr__(self, '_equality_constraints', OrderedSet())
         object.__setattr__(self, '_frequency_constraints', OrderedSet())
+        object.__setattr__(self, '_cardinality_constraints', OrderedSet())
 
     def to_string(self, model: Model, enums_enabled: bool = False, pyi_enabled: bool = False) -> None:
         self._process_constraints(model)
@@ -79,6 +82,7 @@ class Printer(BasePrinter):
         self._print_nl("RoleSubsetConstraint = model.role_subset")
         self._print_nl("EqualityConstraint = model.equality")
         self._print_nl("FrequencyConstraint = model.frequency")
+        self._print_nl("CardinalityConstraint = model.cardinality")
 
         self._nl()
         self._handle_concepts(model, enums_enabled, pyi_enabled)
@@ -133,6 +137,10 @@ class Printer(BasePrinter):
             self._nl()
             self._handle_frequency_constraints()
 
+        if self._cardinality_constraints:
+            self._nl()
+            self._handle_cardinality_constraints()
+
     def _handle_concepts(self, model: Model, enums_enabled: bool, pyi_enabled: bool) -> None:
         concepts_map = model.concepts_map()
         enums_map = model.enums_map()
@@ -177,11 +185,23 @@ class Printer(BasePrinter):
             # print a root Relationship
             self._print_nl(f"{self._get_relationship_name(rel)} = "
                            f"Relationship('{rel._madlib}'{self._print_if_not_empty_and_not_equal('short_name', rel._passed_short_name, rel._name)})")
+            self._handle_role_verbalization(rel)
+
             # print remaining RelationshipReadings if any
             for r in rel._readings[1:]:
                 self._print_nl(f"{self._get_relationship_name(r)} = "
                                f"{self._get_relationship_name(rel)}.alt('{r._madlib}'{self._print_if_not_empty_and_not_equal('short_name', r._passed_short_name, r._name)})")
 
+    def _handle_role_verbalization(self, rel):
+        for ro in rel._roles():
+            verbalization_data = []
+            if ro.prefix:
+                verbalization_data.append(f'prefix="{ro.prefix}"')
+            if ro.postfix:
+                verbalization_data.append(f'postfix="{ro.postfix}"')
+            if verbalization_data:
+                self._print_nl(f"{self._get_role_name(ro)}.verbalization({', '.join(verbalization_data)})")
+
     def _handle_ref_schemes(self, model: Model) -> None:
         for concept_name, concept in model.entity_types_map().items():
             if not isinstance(concept, Concept) or not concept._reference_schemes:
@@ -208,21 +228,7 @@ class Printer(BasePrinter):
     def _handle_role_value_constraints(self) -> None:
         for constraint in self._role_value_constraints:
             role = constraint.roles()[0]
-            elements = []
-
-            for value in constraint.values():
-                if isinstance(value, Range):
-                    if value._start is not None and value._end is not None:
-                        elements.append(f"Range.between({value._start}, {value._end})")
-                    elif value._end is not None:
-                        elements.append(f"Range.to_value({value._end})")
-                    else:
-                        elements.append(f"Range.from_value({value._start})")
-                elif isinstance(value, str):
-                    elements.append(repr(value))
-                else:
-                    elements.append(str(value))
-
+            elements = self._get_constraint_values(constraint)
             role_name = self._get_role_name(role)
             self._print_nl(f"RoleValueConstraint({role_name}, [{', '.join(elements)}])")
 
@@ -267,6 +273,11 @@ class Printer(BasePrinter):
             elements = [self._get_role_name(role) for role in constraint.roles()]
             self._print_nl(f"FrequencyConstraint(({', '.join(constraint.frequency)}), {', '.join(elements)})")
 
+    def _handle_cardinality_constraints(self):
+        for constraint in self._cardinality_constraints:
+            elements = self._get_constraint_values(constraint)
+            self._print_nl(f"CardinalityConstraint({constraint.concept()._name}, [{', '.join(elements)}])")
+
     def _emit_role_sequence_constraints(self, constraints, constraint_type: str):
         for constraint in constraints:
             role_sequences = [[self._get_role_name(r) for r in ro_list] for ro_list in constraint.role_sequences()]
@@ -306,6 +317,8 @@ class Printer(BasePrinter):
                 self._equality_constraints.add(c)
             elif isinstance(c, FrequencyConstraint):
                 self._frequency_constraints.add(c)
+            elif isinstance(c, CardinalityConstraint):
+                self._cardinality_constraints.add(c)
 
     def _get_role_name(self, role: Role) -> str:
         return f"{self._get_relationship_name(role._relationship)}['{role._field_ref._name}']"
@@ -322,6 +335,23 @@ class Printer(BasePrinter):
         name = _get_concept_name(t)
         return f"qb.{name}" if name in Concept.builtins else name
 
+    @staticmethod
+    def _get_constraint_values(constraint):
+        elements = []
+        for value in constraint.values():
+            if isinstance(value, Range):
+                if value._start and value._end:
+                    elements.append(f"Range.between({value._start}, {value._end})")
+                elif value._end:
+                    elements.append(f"Range.to_value({value._end})")
+                else:
+                    elements.append(f"Range.from_value({value._start})")
+            elif isinstance(value, str):
+                elements.append(repr(value))
+            else:
+                elements.append(str(value))
+        return elements
+
     @staticmethod
     def _print_if_not_empty(label: str, value: typing.Optional[str]) -> str:
         return f", {label}='{value}'" if value else ""
