@@ -1,20 +1,21 @@
 from relationalai.early_access.metamodel import ir, builtins as rel_builtins, helpers, types
 from relationalai.early_access.metamodel.visitor import collect_by_type
-from relationalai.early_access.lqp import ir as lqp, utils
+from relationalai.early_access.lqp import ir as lqp, utils, types as lqp_types
 from relationalai.early_access.lqp.primitives import (
     relname_to_lqp_name, lqp_operator, lqp_avg_op, is_monotype
 )
 from relationalai.early_access.lqp.types import meta_type_to_lqp, type_from_constant
 from relationalai.early_access.lqp.constructors import (
-    mk_and, mk_exists, mk_or, mk_abstraction, mk_var
+    mk_and, mk_exists, mk_or, mk_abstraction, mk_value, mk_var, mk_type
 )
 from relationalai.early_access.lqp.utils import TranslationCtx, gen_unique_var
 from relationalai.early_access.lqp.validators import assert_valid_input
 
 from typing import Tuple, cast, Union, Optional
+import re
 
 """ Main access point. Converts the model IR to an LQP transaction. """
-def to_lqp(model: ir.Model, fragment_name: bytes) -> lqp.Transaction:
+def to_lqp(model: ir.Model, fragment_name: bytes) -> tuple[Optional[tuple], lqp.Transaction]:
     assert_valid_input(model)
     ctx = TranslationCtx(model)
     decls: list[lqp.Declaration] = []
@@ -28,10 +29,13 @@ def to_lqp(model: ir.Model, fragment_name: bytes) -> lqp.Transaction:
         decls.extend(new_decls)
 
     reads: list[lqp.Read] = []
-    for (i, (output_id, output_name)) in enumerate(ctx.output_ids):
-        assert isinstance(output_id, lqp.RelationId)
-        output = lqp.Output(name=output_name, relation_id=output_id, meta=None)
-        reads.append(lqp.Read(read_type=output, meta=None))
+    reads.extend(_get_output_reads(ctx.output_ids))
+
+    export_info = None
+    if len(ctx.export_ids) > 0:
+        export_filename, col_types, export_reads = _get_export_reads(ctx.export_ids)
+        reads.extend(export_reads)
+        export_info = (export_filename, col_types)
 
     debug_info = lqp.DebugInfo(id_to_orig_name=ctx.rel_id_to_orig_name, meta=None)
     fragment_id = lqp.FragmentId(id=fragment_name, meta=None)
@@ -42,8 +46,7 @@ def to_lqp(model: ir.Model, fragment_name: bytes) -> lqp.Transaction:
         epochs=[
             lqp.Epoch(
                 reads=reads,
-                local_writes=[lqp.Write(write_type=define_op, meta=None)],
-                persistent_writes=[],
+                writes=[lqp.Write(write_type=define_op, meta=None)],
                 meta=None
             )
         ],
@@ -51,7 +54,7 @@ def to_lqp(model: ir.Model, fragment_name: bytes) -> lqp.Transaction:
     )
 
     lqp.validate_lqp(txn)
-    return txn
+    return (export_info, txn)
 
 def _effect_bindings(effect: Union[ir.Output, ir.Update]) -> list[ir.Value]:
     if isinstance(effect, ir.Output):
@@ -61,6 +64,41 @@ def _effect_bindings(effect: Union[ir.Output, ir.Update]) -> list[ir.Value]:
     else:
         return list(effect.args)
 
+def _get_output_reads(output_ids: list[tuple[lqp.RelationId, str]]) -> list[lqp.Read]:
+    reads = []
+    for (rel_id, name) in output_ids:
+        assert isinstance(rel_id, lqp.RelationId)
+        output = lqp.Output(name=name, relation_id=rel_id, meta=None)
+        reads.append(lqp.Read(read_type=output, meta=None))
+    return reads
+
+def _get_export_reads(export_ids: list[tuple[lqp.RelationId, int, lqp.Type]]) -> tuple[str, list, list[lqp.Read]]:
+    reads = []
+    csv_columns = []
+    col_info = []
+    for (col_id, col_num, col_type) in sorted(export_ids, key=lambda x: x[1]):
+        col_name = f"col{col_num:03}"
+        csv_columns.append(lqp.ExportCSVColumn(column_name=col_name, column_data=col_id, meta=None))
+        col_info.append((col_name, col_type))
+
+    # Generate a hashed name based on relation ids
+    rel_ids = tuple([id[0] for id in export_ids])
+    export_filename = "export_" + str(abs(hash(rel_ids)))
+
+    # Note that the engine will append the transaction id to the export path for access
+    # control reasons. So the actual final path will be
+    # `.../export/{export_filename}/data_{txn_id}`
+    export_path = f"snowflake://APP_STATE.RAI_INTERNAL_STAGE/export/{export_filename}/data"
+    export_csv_config = lqp.ExportCSVConfig(
+        path=export_path,
+        data_columns=csv_columns,
+        compression="gzip",
+        partition_size=200,
+        meta=None,
+    )
+    reads.append(lqp.Read(read_type=lqp.Export(config=export_csv_config, meta=None), meta=None))
+    return (export_filename, col_info, reads)
+
 def _translate_to_decls(ctx: TranslationCtx, rule: ir.Logical) -> list[lqp.Declaration]:
     effects = collect_by_type((ir.Output, ir.Update), rule)
     aggregates = collect_by_type(ir.Aggregate, rule)
@@ -89,12 +127,24 @@ def _translate_to_decls(ctx: TranslationCtx, rule: ir.Logical) -> list[lqp.Decla
 
     return [_translate_effect(ctx, effect, mk_and(conjuncts)) for effect in effects]
 
+# Translates an effect (export, output, or update) into the corresponding def. Note that
+# this method only generates the def and not the LQP read operation, which is added later by
+# `_get_output_reads` and `_get_export_reads`.
 def _translate_effect(ctx: TranslationCtx, effect: Union[ir.Output, ir.Update], body: lqp.Formula) -> lqp.Declaration:
     bindings = _effect_bindings(effect)
+
+    def _is_export(e):
+        return isinstance(e, ir.Output) and rel_builtins.export_annotation in e.annotations
+
     if isinstance(effect, ir.Output):
         projection, eqs, suffix = _translate_output_bindings(ctx, bindings)
         meta_id = effect.id
-        def_name = "output" + suffix
+
+        if _is_export(effect):
+            def_name = "export_relation" + suffix
+        else:
+            def_name = "output" + suffix
+
         # Uniquify the output name
         def_name = ctx.def_names.get_name_by_id(meta_id, def_name)
     else:
@@ -104,10 +154,19 @@ def _translate_effect(ctx: TranslationCtx, effect: Union[ir.Output, ir.Update],
 
     eqs.append(body)
     new_body = mk_and(eqs)
+
     rel_id = get_relation_id(ctx, def_name, meta_id)
 
-    # Context bookkeeping
-    if isinstance(effect, ir.Output):
+    # Context bookkeeping for exports and outputs
+    if _is_export(effect):
+        # The row id is the first n-1 elements, and the actual data is the last element. Its
+        # type is stored in the first element of the tuple.
+        col_type = projection[-1][1]
+        _col_num_match = re.search(r"export_relation_col([0-9]+)", def_name)
+        assert _col_num_match, f"Could not find column number in suffix: {def_name}"
+        col_num = int(_col_num_match.group(1))
+        ctx.export_ids.append((rel_id, col_num, col_type))
+    elif isinstance(effect, ir.Output):
         ctx.output_ids.append((rel_id, def_name))
 
     return lqp.Def(
@@ -117,7 +176,7 @@ def _translate_effect(ctx: TranslationCtx, effect: Union[ir.Output, ir.Update],
         meta = None,
     )
 
-def _translate_output_bindings(ctx: TranslationCtx, bindings: list[ir.Value]) -> Tuple[list[Tuple[lqp.Var, lqp.RelType]], list[lqp.Formula], str]:
+def _translate_output_bindings(ctx: TranslationCtx, bindings: list[ir.Value]) -> Tuple[list[Tuple[lqp.Var, lqp.Type]], list[lqp.Formula], str]:
     symbol_literals = []
     non_symbols = []
     for binding in bindings:
@@ -210,7 +269,7 @@ def _translate_descending_rank(ctx: TranslationCtx, rank: ir.Rank, body: lqp.For
             ctx.var_names,
             "count",
             "count",
-            lqp.PrimitiveType.INT
+            mk_type(lqp.TypeName.INT)
         ),
         body=mk_abstraction(aggr_abstr_args, desc_body),
         terms=[count_res],
@@ -231,9 +290,10 @@ def _translate_descending_rank(ctx: TranslationCtx, rank: ir.Rank, body: lqp.For
     # If there is a limit, we need to add a filter to the ranking.
     # Wrap with a rank <= limit
     if rank.limit != 0:
+        limit_term, _ = _translate_term(ctx, rank.limit)
         limiter = lqp.Primitive(
             name="rel_primitive_lt_eq",
-            terms=[result_var, rank.limit],
+            terms=[result_var, limit_term],
             meta=None
         )
         ranking = mk_and([ranking, limiter])
@@ -293,7 +353,7 @@ def _translate_aggregate(ctx: TranslationCtx, aggr: ir.Aggregate, body: lqp.Form
     projected_args, projected_eqs = _translate_bindings(ctx, list(introduced_meta_projs))
     body_conjs.extend(input_eqs)
     body_conjs.extend(projected_eqs)
-    abstr_args: list[Tuple[lqp.Var, lqp.RelType]] = projected_args + input_args
+    abstr_args: list[Tuple[lqp.Var, lqp.Type]] = projected_args + input_args
 
     if aggr_name == "count" or aggr_name == "avg":
         assert len(output_vars) == 1, "Count and avg expect a single output variable"
@@ -329,7 +389,13 @@ def _translate_aggregate(ctx: TranslationCtx, aggr: ir.Aggregate, body: lqp.Form
 
         # Finally, we need to wrap everything in an `exists` to project away the sum and
         # count variables and only keep the result of the division.
-        result = mk_exists([(sum_result, sum_type), (count_result, lqp.PrimitiveType.INT)], conjunction)
+        result = mk_exists(
+            [
+                (sum_result, sum_type),
+                (count_result, mk_type(lqp.TypeName.INT))
+            ],
+            conjunction,
+        )
 
         return result
 
@@ -384,53 +450,44 @@ def _translate_to_formula(ctx: TranslationCtx, task: ir.Task) -> lqp.Formula:
         raise NotImplementedError(f"Unknown task type (formula): {type(task)}")
 
 # Only used for translating terms on atoms, which can be specialized values.
-def _translate_relterm(ctx: TranslationCtx, term: ir.Value) -> Tuple[lqp.RelTerm, lqp.RelType]:
+def _translate_relterm(ctx: TranslationCtx, term: ir.Value) -> Tuple[lqp.RelTerm, lqp.Type]:
     if isinstance(term, ir.Literal) and term.type == types.Symbol:
         if isinstance(term.value, str):
-            return lqp.Specialized(value=term.value, meta=None), meta_type_to_lqp(types.String)
+            value = lqp.Value(value=term.value, cast_type=None, meta=None)
+            return lqp.SpecializedValue(value=value, meta=None), meta_type_to_lqp(types.String)
         elif isinstance(term.value, int):
-            return lqp.Specialized(value=term.value, meta=None), meta_type_to_lqp(types.Int)
+            value = lqp.Value(value=term.value, cast_type=None, meta=None)
+            return lqp.SpecializedValue(value=value, meta=None), meta_type_to_lqp(types.Int)
         else:
             raise NotImplementedError(f"Cannot specialize literal of type {type(term.value)}")
     return _translate_term(ctx, term)
 
-def _translate_term(ctx: TranslationCtx, term: ir.Value) -> Tuple[lqp.Term, lqp.RelType]:
+def _translate_term(ctx: TranslationCtx, term: ir.Value) -> Tuple[lqp.Term, lqp.Type]:
     if isinstance(term, ir.Var):
         name = ctx.var_names.get_name_by_id(term.id, term.name)
         t = meta_type_to_lqp(term.type)
         return mk_var(name), t
     elif isinstance(term, ir.Literal):
-        assert isinstance(term.value, lqp.PrimitiveValue), f"expected primitive value, got {type(term.value)}: {term.value}"
-        return term.value, meta_type_to_lqp(term.type)
+        assert isinstance(term.value, lqp_types.PrimitiveValue), \
+            f"expected primitive value, got {type(term.value)}: {term.value}"
+        value = lqp.Value(value=term.value, cast_type=None, meta=None)
+        return value, meta_type_to_lqp(term.type)
     else:
-        assert isinstance(term, lqp.PrimitiveValue), \
+        assert isinstance(term, lqp_types.PrimitiveValue), \
             f"Cannot translate value {term!r} of type {type(term)} to LQP Term; not a PrimitiveValue."
-        return term, type_from_constant(term)
-
-# In the metamodel, type conversions are represented as special relations, whereas in LQP we
-# have a dedicated `Cast` node. Eventually we might want to unify these, but for now we use
-# this mapping here.
-rel_to_cast = {
-    "decimal64": lqp.RelValueType.DECIMAL64,
-    "decimal128": lqp.RelValueType.DECIMAL128,
-
-    "int_to_float": lqp.PrimitiveType.FLOAT,
-    "int_to_decimal64": lqp.RelValueType.DECIMAL64,
-    "int_to_decimal128": lqp.RelValueType.DECIMAL128,
-
-    "float_to_int": lqp.PrimitiveType.INT,
-    "float_to_decimal64": lqp.RelValueType.DECIMAL64,
-    "float_to_decimal128": lqp.RelValueType.DECIMAL128,
-
-    "decimal64_to_float": lqp.PrimitiveType.FLOAT,
-    "decimal64_to_int": lqp.PrimitiveType.INT,
-    "decimal64_to_decimal128": lqp.RelValueType.DECIMAL128,
-    "decimal128_to_float": lqp.PrimitiveType.FLOAT,
-    "decimal128_to_int": lqp.PrimitiveType.INT,
-    "decimal128_to_decimal64": lqp.RelValueType.DECIMAL64
-}
+        value = lqp.Value(value=term, cast_type=None, meta=None)
+        return value, type_from_constant(term)
 
 def _translate_to_atom(ctx: TranslationCtx, task: ir.Lookup) -> lqp.Formula:
+    if task.relation == rel_builtins.cast:
+        assert len(task.args) == 3, f"expected three terms for {task.relation.name}, got {len(task.args)}"
+
+        terms = []
+        for arg in task.args[1:]:
+            term, _ = _translate_relterm(ctx, arg)
+            terms.append(term)
+        return lqp.Cast(input=terms[0], result=terms[1], meta=None)
+
     # TODO: want signature not name
     rel_name = task.relation.name
     terms = []
@@ -448,20 +505,51 @@ def _translate_to_atom(ctx: TranslationCtx, task: ir.Lookup) -> lqp.Formula:
             term_types.append(ty)
 
     if rel_builtins.is_builtin(task.relation):
-        if task.relation.name in rel_to_cast:
-            assert len(terms) == 2, f"expected two terms for cast {task.relation.name}, got {terms}"
-            return lqp.Cast(type=rel_to_cast[task.relation.name], input=terms[0], result=terms[1], meta=None)
-        elif task.relation.name == "construct_datetime" and len(terms) == 7:
+        if task.relation.name == "construct_datetime" and len(terms) == 7:
             # construct_datetime does not provide a timezone or milliseconds so we
             # default to 0 milliseconds and UTC timezone.
             lqp_name = relname_to_lqp_name(task.relation.name)
-            extended_terms = [*terms[:-1], 0, "UTC", terms[-1]]
+            extended_terms = [
+                *terms[:-1],
+                mk_value(0),
+                mk_value("UTC"),
+                terms[-1],
+            ]
             return lqp.Primitive(name=lqp_name, terms=extended_terms, meta=None)
         elif task.relation.name == "like_match":
             # In the like_match API of QB, the variable is the first argument and the pattern the second,
             #   but in Rel, the expected argument order is reversed: pattern first, then variable.
             lqp_name = relname_to_lqp_name(task.relation.name)
             return lqp.Primitive(name=lqp_name, terms=terms[::-1], meta=None)
+        elif task.relation.name == "parse_decimal64" and len(terms) == 2:
+            # If the builtin for parsing decimals is called directly we will need to add
+            # the default parameters of 64 bits and precision 6.
+            lqp_name = relname_to_lqp_name(task.relation.name)
+            bit_value = lqp.Value(value=64, cast_type=None, meta=None)
+            precision_value = lqp.Value(value=6, cast_type=None, meta=None)
+            extended_terms = [
+                lqp.SpecializedValue(value=bit_value, meta=None),
+                lqp.SpecializedValue(value=precision_value, meta=None),
+                terms[0],
+                terms[1],
+            ]
+            return lqp.Primitive(name=lqp_name, terms=extended_terms, meta=None)
+        elif (
+            task.relation.name == "parse_decimal" or
+            task.relation.name == "parse_decimal128"
+         ) and len(terms) == 2:
+            # If the builtin for parsing decimals is called directly we will need to add
+            # the default parameters of 128 bits and precision 10.
+            lqp_name = relname_to_lqp_name(task.relation.name)
+            bit_value = lqp.Value(value=128, cast_type=None, meta=None)
+            precision_value = lqp.Value(value=10, cast_type=None, meta=None)
+            extended_terms = [
+                lqp.SpecializedValue(value=bit_value, meta=None),
+                lqp.SpecializedValue(value=precision_value, meta=None),
+                terms[0],
+                terms[1],
+            ]
+            return lqp.Primitive(name=lqp_name, terms=extended_terms, meta=None)
         else:
             lqp_name = relname_to_lqp_name(task.relation.name)
             if is_monotype(task.relation.name):
@@ -485,7 +573,7 @@ def get_relation_id(ctx: TranslationCtx, orig_name: str, metamodel_id: int) -> l
     ctx.rel_id_to_orig_name[relation_id] = unique_name
     return relation_id
 
-def _translate_bindings(ctx: TranslationCtx, bindings: list[ir.Value]) -> Tuple[list[Tuple[lqp.Var, lqp.RelType]], list[lqp.Formula]]:
+def _translate_bindings(ctx: TranslationCtx, bindings: list[ir.Value]) -> Tuple[list[Tuple[lqp.Var, lqp.Type]], list[lqp.Formula]]:
     lqp_vars = []
     conjuncts = []
     for binding in bindings:
@@ -496,24 +584,29 @@ def _translate_bindings(ctx: TranslationCtx, bindings: list[ir.Value]) -> Tuple[
 
     return lqp_vars, conjuncts
 
-def binding_to_lqp_var(ctx: TranslationCtx, binding: ir.Value) -> Tuple[lqp.Var, lqp.RelType, Union[None, lqp.Formula]]:
+def binding_to_lqp_var(ctx: TranslationCtx, binding: ir.Value) -> Tuple[lqp.Var, lqp.Type, Union[None, lqp.Formula]]:
     if isinstance(binding, ir.Var):
         var, typ = _translate_term(ctx, binding)
         assert isinstance(var, lqp.Var)
         return var, typ, None
     elif isinstance(binding, ir.Literal):
         return constant_to_lqp_var(ctx, binding.value, binding.type)
-    elif isinstance(binding, lqp.PrimitiveValue):
+    elif isinstance(binding, lqp_types.PrimitiveValue):
         return constant_to_lqp_var(ctx, binding, None)
     else:
         raise Exception(f"Unsupported binding type: {type(binding)}")
 
-def constant_to_lqp_var(ctx: TranslationCtx, value: lqp.PrimitiveValue, type: Optional[ir.Type], name_hint: str = "cvar") -> Tuple[lqp.Var, lqp.RelType, lqp.Formula]:
+def constant_to_lqp_var(
+    ctx: TranslationCtx,
+    value: lqp_types.PrimitiveValue,
+    type: Optional[ir.Type], name_hint: str = "cvar"
+) -> Tuple[lqp.Var, lqp.Type, lqp.Formula]:
     var = gen_unique_var(ctx, name_hint)
     if type is not None:
         typ = meta_type_to_lqp(type)
     else:
         typ = type_from_constant(value)
-    eq = lqp.Primitive(name="rel_primitive_eq", terms=[var, value], meta=None)
+    value_term = lqp.Value(value=value, cast_type=None, meta=None)
+    eq = lqp.Primitive(name="rel_primitive_eq", terms=[var, value_term], meta=None)
 
     return var, typ, eq
