@@ -1,9 +1,11 @@
 from relationalai.early_access.lqp import ir as lqp
 from relationalai.early_access.lqp.utils import UniqueNames
+from relationalai.early_access.lqp.constructors import mk_type
 
 rel_to_lqp = {
     "=": "rel_primitive_eq",
     "!=": "rel_primitive_neq",
+    "%": "rel_primitive_remainder",
     "abs": "rel_primitive_abs",
     "construct_date": "rel_primitive_construct_date",
     "construct_datetime": "rel_primitive_construct_datetime",
@@ -27,6 +29,8 @@ rel_to_lqp = {
     "date_month": "rel_primitive_date_month",
     "date_day": "rel_primitive_date_day",
     "range": "rel_primitive_range",
+    "natural_log": "rel_primitive_natural_log",
+    "sqrt": "rel_primitive_sqrt",
     # Division is monotype, but only on the input args. Until we distinguish between input
     # and output args, we can't use the same assertions for monotype-ness as the other ops.
     "/": "rel_primitive_divide_monotype",
@@ -48,6 +52,7 @@ rel_to_lqp_monotype = {
     ">=": "rel_primitive_gt_eq_monotype",
     ">": "rel_primitive_gt_monotype",
     "<": "rel_primitive_lt_monotype",
+    "maximum": "rel_primitive_max",
 }
 
 def relname_to_lqp_name(name: str) -> str:
@@ -65,8 +70,8 @@ def is_monotype(name: str) -> bool:
 
 # We take the name and type of the variable that we're summing over, so that we can generate
 # recognizable names for the variables in the reduce operation and preserve the type.
-def lqp_avg_op(names: UniqueNames, op_name: str, sum_name: str, sum_type: lqp.RelType) -> lqp.Abstraction:
-    count_type = lqp.PrimitiveType.INT
+def lqp_avg_op(names: UniqueNames, op_name: str, sum_name: str, sum_type: lqp.Type) -> lqp.Abstraction:
+    count_type = mk_type(lqp.TypeName.INT)
     vars = [
         (lqp.Var(name=names.get_name(sum_name), meta=None), sum_type),
         (lqp.Var(name=names.get_name("counter"), meta=None), count_type),
@@ -93,7 +98,7 @@ def lqp_avg_op(names: UniqueNames, op_name: str, sum_name: str, sum_type: lqp.Re
     return lqp.Abstraction(vars=vars, value=body, meta=None)
 
 # Default handler for aggregation operations in LQP.
-def lqp_agg_op(names: UniqueNames, op_name: str, aggr_arg_name: str, aggr_arg_type: lqp.RelType) -> lqp.Abstraction:
+def lqp_agg_op(names: UniqueNames, op_name: str, aggr_arg_name: str, aggr_arg_type: lqp.Type) -> lqp.Abstraction:
     x = lqp.Var(name=names.get_name(f"x_{aggr_arg_name}"), meta=None)
     y = lqp.Var(name=names.get_name(f"y_{aggr_arg_name}"), meta=None)
     z = lqp.Var(name=names.get_name(f"z_{aggr_arg_name}"), meta=None)
@@ -104,7 +109,7 @@ def lqp_agg_op(names: UniqueNames, op_name: str, aggr_arg_name: str, aggr_arg_ty
 
     return lqp.Abstraction(vars=ts, value=body, meta=None)
 
-def lqp_operator(names: UniqueNames, op_name: str, aggr_arg_name: str, aggr_arg_type: lqp.RelType) -> lqp.Abstraction:
+def lqp_operator(names: UniqueNames, op_name: str, aggr_arg_name: str, aggr_arg_type: lqp.Type) -> lqp.Abstraction:
     # TODO: Can we just pass through unknown operations?
     if op_name not in agg_to_lqp:
         raise NotImplementedError(f"Unsupported aggregation: {op_name}")
