@@ -109,6 +109,20 @@ maximum = f.relation(
     ],
 )
 
+minimum = f.relation(
+    "minimum",
+    [f.input_field("a", types.Number), f.input_field("b", types.Number), f.field("c", types.Number)],
+    overloads=[
+        f.relation("minimum", [f.input_field("a", types.Int), f.input_field("b", types.Int), f.field("c", types.Int)]),
+        f.relation("minimum", [f.input_field("a", types.Float), f.input_field("b", types.Float), f.field("c", types.Float)]),
+        f.relation("minimum", [f.input_field("a", types.Decimal64), f.input_field("b", types.Decimal64), f.field("c", types.Decimal64)]),
+        f.relation("minimum", [f.input_field("a", types.Decimal128), f.input_field("b", types.Decimal128), f.field("c", types.Decimal128)]),
+    ],
+)
+
+isinf = f.relation("isinf", [f.input_field("a", types.Float)])
+isnan = f.relation("isnan", [f.input_field("a", types.Float)])
+
 # Strings
 concat = f.relation("concat", [f.input_field("a", types.String), f.input_field("b", types.String), f.field("c", types.String)])
 num_chars = f.relation("num_chars", [f.input_field("a", types.String), f.field("b", types.Int)])
@@ -126,20 +140,6 @@ date_year = f.relation("date_year", [f.input_field("a", types.Date), f.field("b"
 date_month = f.relation("date_month", [f.input_field("a", types.Date), f.field("b", types.Int)])
 date_day = f.relation("date_day", [f.input_field("a", types.Date), f.field("b", types.Int)])
 
-# Conversions
-int_to_float = f.relation("int_to_float", [f.input_field("a", types.Int), f.field("b", types.Float)])
-int_to_decimal64 = f.relation("int_to_decimal64", [f.input_field("a", types.Int), f.field("b", types.Decimal64)])
-int_to_decimal128 = f.relation("int_to_decimal128", [f.input_field("a", types.Int), f.field("b", types.Decimal128)])
-float_to_int = f.relation("float_to_int", [f.input_field("a", types.Float), f.field("b", types.Int)])
-float_to_decimal64 = f.relation("float_to_decimal64", [f.input_field("a", types.Float), f.field("b", types.Decimal64)])
-float_to_decimal128 = f.relation("float_to_decimal128", [f.input_field("a", types.Float), f.field("b", types.Decimal128)])
-decimal64_to_int = f.relation("decimal64_to_int", [f.input_field("a", types.Decimal64), f.field("b", types.Int)])
-decimal64_to_float = f.relation("decimal64_to_float", [f.input_field("a", types.Decimal64), f.field("b", types.Float)])
-decimal64_to_decimal128 = f.relation("decimal64_to_decimal128", [f.input_field("a", types.Decimal64), f.field("b", types.Decimal128)])
-decimal128_to_int = f.relation("decimal128_to_int", [f.input_field("a", types.Decimal128), f.field("b", types.Int)])
-decimal128_to_float = f.relation("decimal128_to_float", [f.input_field("a", types.Decimal128), f.field("b", types.Float)])
-decimal128_to_decimal64 = f.relation("decimal128_to_decimal64", [f.input_field("a", types.Decimal128), f.field("b", types.Decimal64)])
-
 # Other
 range = f.relation("range", [
     f.input_field("start", types.Int),
@@ -177,9 +177,13 @@ concept_relation_annotation = f.annotation(concept_population, [])
 from_cdc = f.relation("from_cdc", [])
 from_cdc_annotation = f.annotation(from_cdc, [])
 
-# indicates this relation is a type_box operation for value types
-type_box = f.relation("type_box", [])
-type_box_annotation = f.annotation(type_box, [])
+# indicates an = lookup that is from a cast operation for value types
+from_cast = f.relation("from_cast", [])
+from_cast_annotation = f.annotation(from_cast, [])
+
+# indicates the original keys of an output (before they were replaced by a compound key)
+output_keys = f.relation("output_keys", [])
+output_keys_annotation = f.annotation(output_keys, [])
 
 #
 # Aggregations
@@ -269,11 +273,14 @@ parse_decimal128 = f.relation("parse_decimal128", [f.input_field("a", types.Stri
 # Alias parse_decimal128 to parse_decimal
 parse_decimal = parse_decimal128
 
-# Implicit decimal conversions
-# These allow writing decimal128(1) in QB and letting the typer insert the conversion from any number type
-decimal64 = f.relation("decimal64", [f.input_field("input", types.Decimal64), f.field("result", types.Decimal64)])
-decimal128 = f.relation("decimal128", [f.input_field("input", types.Decimal128), f.field("result", types.Decimal128)])
-
+cast = f.relation(
+    "cast",
+    [
+        f.input_field("to_type", types.Any),
+        f.input_field("source",  types.Any),
+        f.field("target",        types.Any)
+    ]
+)
 
 # Date construction with less overhead
 construct_date = f.relation("construct_date", [f.input_field("year", types.Int), f.input_field("month", types.Int), f.input_field("day", types.Int), f.field("date", types.Date)])
@@ -326,17 +333,16 @@ def _compute_builtin_overloads() -> list[ir.Relation]:
     return overloads
 
 # manually maintain the list of relations that are actually annotations
-builtin_annotations = [external, export, concept_population, from_cdc, type_box]
+builtin_annotations = [external, export, concept_population, from_cdc, from_cast]
 builtin_annotations_by_name = dict((r.name, r) for r in builtin_annotations)
 
 builtin_relations = _compute_builtin_relations()
 builtin_overloads = _compute_builtin_overloads()
 builtin_relations_by_name = dict((r.name, r) for r in builtin_relations)
 
-conversion_builtins = [string, decimal64, decimal128, int_to_float, int_to_decimal64, int_to_decimal128, float_to_int,
-                              float_to_decimal64, float_to_decimal128, decimal64_to_int, decimal64_to_float,
-                              decimal64_to_decimal128, decimal128_to_int, decimal128_to_float, decimal128_to_decimal64]
-
 string_binary_builtins = [num_chars, starts_with, ends_with, contains, like_match, lower, upper]
 
 date_builtins = [date_year, date_month, date_day]
+
+math_builtins = [abs, *abs.overloads, natural_log, *natural_log.overloads, sqrt, *sqrt.overloads,
+                 maximum, *maximum.overloads, minimum, *minimum.overloads]
