@@ -2,5 +2,6 @@ from .splinter import Splinter
 from .list_types import RewriteListTypes
 from .gc_nodes import GarbageCollectNodes
 from .flatten import Flatten
+from .extract_keys import ExtractKeys
 
-__all__ = ["Splinter", "RewriteListTypes", "GarbageCollectNodes", "Flatten"]
+__all__ = ["Splinter", "RewriteListTypes", "GarbageCollectNodes", "Flatten", "ExtractKeys"]
