@@ -127,12 +127,12 @@ class Flatten(Pass):
 
     def __init__(self, handle_outputs: bool=True):
         super().__init__()
-        self.name_cache = NameCache()
+        self.name_cache = NameCache(start_from_one=True)
         self._handle_outputs = handle_outputs
 
     def create_task_name(self, task: ir.Task, prefix: Optional[str]=None) -> str:
         prefix = prefix if prefix else f"_{task.kind}"
-        return self.name_cache.get_name(task.id, prefix, start_from_one=True)
+        return self.name_cache.get_name(task.id, prefix)
 
     #--------------------------------------------------
     # Public API
@@ -502,6 +502,17 @@ class Flatten(Pass):
                 deps = info.task_dependencies(output)
                 # TODO: verify safety of doing this
                 # self.remove_subsumptions(deps, ctx)
+
+                # add an explicit check for null values in the last argument of the output
+                # null values might be introduced there if the column value is also a nullable
+                # key, but we don't want to explicitly represent them in the column values
+                assert isinstance(output, ir.Output)
+                output_keys_anno = next(filter(lambda anno: anno.relation == builtins.output_keys, output.annotations), None)
+                if output_keys_anno:
+                    last_arg = output.aliases[-1]
+                    if last_arg and isinstance(last_arg[1], ir.Var) and last_arg[1] in output_keys_anno.args:
+                        deps.add(f.lookup(builtins.neq, [last_arg[1], None]))
+
                 deps.add(output)
                 ctx.rewrite_ctx.top_level.append(ir.Logical(task.engine, tuple(), tuple(deps)))
 
