@@ -103,14 +103,6 @@ class CheckModel(visitor.DAGVisitor):
         for t in model.types:
             t.accept(CheckPresence(model, env))
 
-        # Check that scalar type names are unique.
-        type_names = set()
-        for t in model.types:
-            if isinstance(t, ir.ScalarType):
-                if t.name in type_names:
-                    env.diags.append(CheckError(f"Type {t} has is not unique {t.name}.", t))
-                type_names.add(t.name)
-
         # Check that the types are acyclic and that the supertype relation is acyclic.
         def check_type_cycles(t: ir.Type, subtypes: OrderedSet[ir.Type]=ordered_set(), containers: OrderedSet[ir.Type]=ordered_set()) -> None:
             if t in subtypes:
