@@ -9,6 +9,7 @@ from relationalai import debugging
 from relationalai.early_access.metamodel import builtins, helpers, ir, types, visitor, compiler, factory as f
 from relationalai.early_access.metamodel.util import OrderedSet, ordered_set
 import rich
+import sys
 
 #--------------------------------------------------
 # Helpers
@@ -43,15 +44,15 @@ TYPE_MAPPING = {
 }
 
 # map potential type conversions to their cost
-CONVERSION_LATTICE:dict[tuple[ir.Type|None, ir.Type|None], float] = {
+CONVERSION_LATTICE:set[tuple[ir.Type, ir.Type]] = {
     # Int -> Decimal64 -> Decimal128 -> Float
-    (types.Int, types.Decimal64): 3.0,
-    (types.Decimal64, types.Decimal128): 5.0,
-    (types.Decimal128, types.Float): 7.0,
+    (types.Int, types.Decimal64),
+    (types.Decimal64, types.Decimal128),
+    (types.Decimal128, types.Float),
     # transitive closure of the above
-    (types.Int, types.Decimal128): 8.0,
-    (types.Decimal64, types.Float): 12.0,
-    (types.Int, types.Float): 15.0,
+    (types.Int, types.Decimal128),
+    (types.Decimal64, types.Float),
+    (types.Int, types.Float),
 }
 
 PRIMITIVES = [types.Int, types.Float, types.Decimal64, types.Decimal128, types.Bool, types.String, types.Date, types.DateTime]
@@ -123,6 +124,15 @@ def type_matches(actual:ir.Type, expected:ir.Type, allow_expected_parents=False)
 def conversion_allowed(arg:ir.Value, actual:ir.Type, expected:ir.Type, arg_types:set[ir.Type]=set()) -> bool:
     return conversion_cost(arg, actual, expected, arg_types) < float("inf")
 
+def literal_conversion_allowed(actual:ir.Type, expected:ir.Type) -> bool:
+    if actual == types.Int and expected in (types.Float, types.Decimal64, types.Decimal128):
+        return True
+    elif actual == types.Float and expected in (types.Decimal64, types.Decimal128):
+        return True
+    elif actual in (types.Decimal64, types.Decimal128) and expected in (types.Decimal64, types.Decimal128):
+        return True
+    return False
+
 def conversion_cost(arg: ir.Value, actual:ir.Type, expected:ir.Type, arg_types:set[ir.Type]) -> float:
     if type_matches(actual, expected):
         return 0
@@ -132,22 +142,33 @@ def conversion_cost(arg: ir.Value, actual:ir.Type, expected:ir.Type, arg_types:s
     elif expected is types.EntityTypeVar and not is_primitive(actual) and len(arg_types) == 1:
         return 0
 
-    # Literal numbers can be converted to other number types.
-    if isinstance(arg, ir.Literal) or isinstance(arg, (int, float, PyDecimal)):
-        if actual == types.Int and expected in (types.Float, types.Decimal64, types.Decimal128):
-            return 1.0
-        elif actual == types.Float and expected in (types.Decimal64, types.Decimal128):
-            return 1.0
-        elif actual in (types.Decimal64, types.Decimal128) and expected in (types.Decimal64, types.Decimal128):
-            return 1.0
+    # Literal numbers can be converted to other number types (but not to value types).
+    if isinstance(arg, ir.Literal) or isinstance(arg, types.py_literal_types):
+        if literal_conversion_allowed(actual, expected):
+            return 1
 
     base_actual = to_base_primitive(actual)
     base_expected = to_base_primitive(expected)
+
+    if not base_actual or not base_expected:
+        return float("inf")
+
     if (base_actual, base_expected) in CONVERSION_LATTICE:
-        return CONVERSION_LATTICE[(base_actual, base_expected)]
+        # The cost for converting from one type to another should be the same regardless of the types involved.
+        # Otherwise, we'll introduce spurious conversions of say Int->Decimal64 plus a cheap literal conversion
+        # Float->Decimal64, rather than a single literal conversion Int->Float.
+        # Use a large enough number here so that if all arguments are literals, it's cheaper to convert
+        # the literals (cost 1, above) rather than not convert some literals, but convert a non-literal instead.
+        # For instance, for `y=1+x` is should be cheaper to convert 1 to float and not convert x and y, than it would be
+        # convert both x and y to integer.
+        return 100
+
     return float("inf")
 
-def merge_types(type1: ir.Type, type2: ir.Type) -> ir.Type:
+def merge_types(type1: ir.Type, type2: ir.Type, is_literal1:bool, is_literal2:bool) -> ir.Type:
+    if type1 == type2:
+        return type1
+
     combined = ordered_set()
     types_to_process = [type1, type2]
 
@@ -163,6 +184,14 @@ def merge_types(type1: ir.Type, type2: ir.Type) -> ir.Type:
     if len(combined) > 1 and types.Any in combined:
         combined.remove(types.Any)
 
+    # If we have multiple types and one is a literal and the other is not,
+    # we can try to convert the literal to the other type.
+    if len(combined) > 1:
+        if is_literal1 and not is_literal2 and literal_conversion_allowed(type1, type2):
+            return type2
+        if is_literal2 and not is_literal1 and literal_conversion_allowed(type2, type1):
+            return type1
+
     # Return single type or create a union
     return next(iter(combined)) if len(combined) == 1 else f.union_type(list(combined))
 
@@ -212,6 +241,10 @@ def try_preserve_type(types:set[ir.Type]) -> Optional[ir.Type]:
                 return t1
     return None
 
+# Was this variable created for a shared literal?
+def var_is_from_literal(var: ir.Var) -> bool:
+    return var.name.startswith("__literal__")
+
 #--------------------------------------------------
 # Type Errors
 #--------------------------------------------------
@@ -280,6 +313,7 @@ class PropagationNetwork():
         self.errors:list[TyperError] = []
         self.resolved_types:dict[Node, ir.Type] = {}
         self.resolved_overloads:dict[int, list[ir.Relation]] = {}
+        self.node_is_literal:dict[Node, bool] = {}
 
     #--------------------------------------------------
     # Mismatches
@@ -315,12 +349,16 @@ class PropagationNetwork():
         self.edges[source].add(target)
         self.has_incoming.add(target)
 
-    def add_type(self, node:Node, type:ir.Type):
+    def add_type(self, node:Node, type:ir.Type, is_literal:bool):
         if node in self.resolved_types:
-            self.resolved_types[node] = merge_types(self.resolved_types[node], type)
+            self.resolved_types[node] = merge_types(self.resolved_types[node], type, self.node_is_literal[node], is_literal)
         else:
             self.resolved_types[node] = type
 
+        if node in self.node_is_literal:
+            is_literal &= self.node_is_literal[node]
+        self.node_is_literal[node] = is_literal
+
     def add_type_requirement(self, source:Node, field:ir.Field):
         self.type_requirements[source].add(field)
 
@@ -331,7 +369,7 @@ class PropagationNetwork():
     def load_types(self, type_dict:dict[Node, ir.Type]):
         for node, type in type_dict.items():
             if isinstance(node, (ir.Field)):
-                self.add_type(node, type)
+                self.add_type(node, type, False)
                 self.loaded_roots.add(node)
                 self.roots.add(node)
 
@@ -414,7 +452,6 @@ class PropagationNetwork():
                     min_cost = total
                     matches.clear()
                 matches.append(overload)
-
         return matches
 
     #--------------------------------------------------
@@ -433,14 +470,20 @@ class PropagationNetwork():
                 continue
             node_type = self.resolve(node)
             if not is_abstract(node_type):
+                is_literal = isinstance(node, ir.Var) and var_is_from_literal(node)
+                self.add_type(node, node_type, is_literal)
                 work_list.append(node)
             else:
                 unhandled.add(node)
 
+        # We need to visit nodes in topological order; that is we need to visit all predecessors
+        # of a node before the node itself.
+
         # push known type nodes through the edges
         for source in work_list:
             unhandled.remove(source)
             source_type = self.resolve(source)
+            is_literal = self.node_is_literal.get(source, False)
             # check to see if the source has ended up with a set of types that
             # aren't valid, e.g. a union of primitives
             if invalid_type(source_type):
@@ -461,7 +504,7 @@ class PropagationNetwork():
                 # otherwise, we just add to the outgoing node's type and if it
                 # changes we add it to the work list
                 elif start := self.resolve(out):
-                    self.add_type(out, source_type)
+                    self.add_type(out, source_type, is_literal)
                     if start != self.resolve(out) or out in unhandled:
                         work_list.append(out)
 
@@ -477,9 +520,7 @@ class PropagationNetwork():
                 if not type_matches(node_type, field_type):
                     node_base = to_base_primitive(node_type)
                     field_base = to_base_primitive(field_type)
-                    if (node_base, field_base) in CONVERSION_LATTICE:
-                        pass
-                    else:
+                    if (node_base, field_base) not in CONVERSION_LATTICE:
                         self.mismatch(node, field_type, node_type)
 
         return self.resolved_types
@@ -509,7 +550,8 @@ class PropagationNetwork():
                                         for field, field_type in zip(overload.fields, resolved_fields)]
             for field_type, arg_type, arg in zip(resolved_fields, resolved_args, node.args):
                 if is_abstract(arg_type) and isinstance(arg, ir.Var):
-                    self.add_type(arg, field_type)
+                    is_literal = var_is_from_literal(arg)
+                    self.add_type(arg, field_type, is_literal)
 
 
 
@@ -565,7 +607,7 @@ class PropagationNetwork():
                 ir_type = resolved.get(node) or self.resolve(node)
                 type_str = type_span(ir_type)
                 # if this is one of our vars that is a placeholder for a literal, just use the type
-                if isinstance(ir_type, ir.ScalarType) and node.name == ir_type.name:
+                if isinstance(ir_type, ir.ScalarType) and var_is_from_literal(node):
                     label = f'(["Literal {type_str}"])'
                 else:
                     label = f'(["{node.name}: {type_str}"])'
@@ -649,7 +691,7 @@ class PropagationNetwork():
             info = ""
             if isinstance(node, ir.Var):
                 ir_type = resolved.get(node) or self.resolve(node)
-                if not(isinstance(ir_type, ir.ScalarType) and node.name == ir_type.name):
+                if not(isinstance(ir_type, ir.ScalarType) and var_is_from_literal(node)):
                     info = f'{node.name}|{type_to_str(ir_type)}'
             elif isinstance(node, ir.Field):
                 ir_type = resolved.get(node) or self.resolve(node)
@@ -690,7 +732,7 @@ class Analyzer(visitor.Visitor):
         # the graph. This allows us to collapse all literals into a single node
         # and avoids exploding node count if lots of data gets added directly
         self.shared_literal_vars:dict[ir.Type, ir.Var] = {
-            t: f.var(t.name, t)
+            t: f.var(f"__literal__{t.name}", t)
             for t in types.builtin_types
             if isinstance(t, ir.ScalarType)
         }
@@ -720,6 +762,8 @@ class Analyzer(visitor.Visitor):
         rel = to_relation(node)
         if isinstance(node, ir.Lookup) and rel == builtins.eq:
             return self.visit_eq(node, parent)
+        if isinstance(node, ir.Lookup) and rel == builtins.cast:
+            return self.visit_cast(node, parent)
 
         if rel.overloads:
             return self.visit_overloaded(node, parent)
@@ -731,7 +775,7 @@ class Analyzer(visitor.Visitor):
         for arg, field in zip(node.args, rel.fields):
             field_type = field.type
             arg_type = self.net.resolve(arg)
-            if not type_matches(arg_type, field_type, allow_any_parents):
+            if not type_matches(arg_type, field_type, allow_any_parents or is_abstract(arg_type)):
                 # Do not complain if we can convert the arg to the field type.
                 if not conversion_allowed(arg, arg_type, field_type, arg_types):
                     self.net.mismatch(node, field_type, arg_type)
@@ -783,6 +827,28 @@ class Analyzer(visitor.Visitor):
             if not conversion_allowed(left, left_type, right_type) and not conversion_allowed(right, right_type, left_type):
                 self.net.mismatch(node, left_type, right_type)
 
+    #--------------------------------------------------
+    # Cast
+    #--------------------------------------------------
+
+    def visit_cast(self, node: ir.Lookup, parent: Optional[ir.Node]):
+        # Cast has fields: to_type (input), source (input), target (output)
+        assert len(node.args) == 3, f"Expected 3 arguments for cast builtin, but got: {node.args}"
+        (to_type, source, target) = node.args
+        assert isinstance(to_type, ir.Type), f"Invalid target type for cast: {to_type}"
+        source_type = self.net.resolve(source)
+
+        if not is_abstract(to_type):
+            # Check if we can convert from the source_type to the to_type.
+            if not is_abstract(source_type):
+                if not conversion_allowed(source, source_type, to_type):
+                    self.net.mismatch(node, source_type, to_type)
+
+            # If target is abstract, it should get the type from to_type
+            target_type = self.net.resolve(target)
+            if is_abstract(target_type) and isinstance(target, ir.Var):
+                self.net.add_type(target, to_type, False)
+
     #--------------------------------------------------
     # Update
     #--------------------------------------------------
@@ -805,9 +871,8 @@ class Analyzer(visitor.Visitor):
 
             # if the field is abstract, then eventually this var will help determine
             # the field's type
-            if is_abstract(field_type):
+            elif is_abstract(field_type):
                 source = self.shared_literal_var(arg_type) if not is_var else arg
-                assert source
                 self.net.add_edge(source, field)
 
             elif not type_matches(arg_type, field_type, allow_expected_parents=allow_any_parents):
@@ -818,21 +883,6 @@ class Analyzer(visitor.Visitor):
 # Replacer
 #--------------------------------------------------
 
-CONVERSIONS:dict[tuple[ir.Type|None, ir.Type|None], ir.Relation] = {
-    (types.Int, types.Float): builtins.int_to_float,
-    (types.Int, types.Decimal64): builtins.int_to_decimal64,
-    (types.Int, types.Decimal128): builtins.int_to_decimal128,
-    (types.Float, types.Int): builtins.float_to_int,
-    (types.Float, types.Decimal64): builtins.float_to_decimal64,
-    (types.Float, types.Decimal128): builtins.float_to_decimal128,
-    (types.Decimal64, types.Int): builtins.decimal64_to_int,
-    (types.Decimal64, types.Float): builtins.decimal64_to_float,
-    (types.Decimal64, types.Decimal128): builtins.decimal64_to_decimal128,
-    (types.Decimal128, types.Int): builtins.decimal128_to_int,
-    (types.Decimal128, types.Float): builtins.decimal128_to_float,
-    (types.Decimal128, types.Decimal64): builtins.decimal128_to_decimal64,
-}
-
 # Once we've pushed all the types through the network, we need to
 # to replace the types of fields and vars that we may have discovered.
 # We also need to replace overloaded lookups with the choosen overloads
@@ -858,14 +908,13 @@ class Replacer(visitor.Rewriter):
             return f.var(node.name, self.net.resolved_types[node])
         return node
 
-    def handle_lookup(self, node: ir.Lookup, parent: ir.Node):
-        # Boxing lookups are present to correctly guide type inference.
-        # They should not propagate further down the pipeline.
-        # Rewrite "type_box(from, to)" to "eq(from, to)"
-        if helpers.is_type_box(node):
-            annos = node.relation.annotations
-            return f.lookup(builtins.eq, [node.args[0], node.args[1]], annos=list(annos))
+    def handle_literal(self, node: ir.Literal, parent: ir.Node):
+        # Up until now, we allow literals to have a non-builtin types. This code converts the value
+        # to a builtin Python type and lifts the literal to that supertype.
+        v, t = self.convert_literal_value(node.value, node.type)
+        return node.reconstruct(value=v, type=t)
 
+    def handle_lookup(self, node: ir.Lookup, parent: ir.Node):
         node = super().handle_lookup(node, parent)
         lookup_args = node.args
 
@@ -884,31 +933,11 @@ class Replacer(visitor.Rewriter):
             right_type = to_type(right)
             mismatch = False
             if not type_matches(left_type, right_type):
-                left_base = to_base_primitive(left_type)
-                right_base = to_base_primitive(right_type)
                 if conversion_allowed(left, left_type, right_type):
-                    name = f"v{to_name(right_type)}"
-                    if isinstance(left, ir.Var):
-                        name = left.name + "_" + to_name(right_type)
-                    new_left = f.var(name, right_type)
-                    conv = CONVERSIONS[(left_base, right_base)]
-                    self.new_relations.add(conv)
-                    tasks.append(f.lookup(
-                        conv,
-                        [left, new_left]
-                    ))
+                    new_left = self.convert(left, left_type, right_type, tasks)
                     final_args = [new_left, right]
                 elif conversion_allowed(right, right_type, left_type):
-                    name = f"v{to_name(left_type)}"
-                    if isinstance(right, ir.Var):
-                        name = right.name + "_" + to_name(left_type)
-                    new_right = f.var(name, left_type)
-                    conv = CONVERSIONS[(right_base, left_base)]
-                    self.new_relations.add(conv)
-                    tasks.append(f.lookup(
-                        conv,
-                        [right, new_right]
-                    ))
+                    new_right = self.convert(right, right_type, left_type, tasks)
                     final_args = [left, new_right]
                 else:
                     self.net.mismatch(node, left_type, right_type)
@@ -944,6 +973,27 @@ class Replacer(visitor.Rewriter):
                 tasks.append(f.lookup(node.relation, final_args))
 
             branches.append(f.logical(tasks, hoists))
+        elif node.relation == builtins.cast:
+            assert len(node.args) == 3, f"Invalid number of arguments for cast: {node.args}"
+            (tgt_type, src, tgt) = node.args
+            assert isinstance(tgt_type, ir.Type), f"Invalid target type for cast: {tgt_type}"
+            src_type = to_type(src)
+
+            src_base = to_base_primitive(src_type)
+            tgt_base = to_base_primitive(tgt_type)
+
+            if src_base and tgt_base and src_base == tgt_base:
+                annos = node.annotations
+                return f.lookup(builtins.eq, (src, tgt), annos=(list(annos) + [builtins.from_cast_annotation]))
+
+            if conversion_allowed(node.args[0], src_type, tgt_type):
+                if tgt_base:
+                    return node.reconstruct(args=(tgt_base, src, tgt))
+                else:
+                    return node
+            else:
+                self.net.mismatch(node, src_type, tgt_type)
+                return node
         else:
             if node.id in self.net.resolved_overloads:
                 resolved = self.net.resolved_overloads[node.id]
@@ -964,17 +1014,8 @@ class Replacer(visitor.Rewriter):
                     if field_type != types.EntityTypeVar and not type_matches(arg_type, field_type):
                         arg_base = to_base_primitive(arg_type)
                         field_base = to_base_primitive(field_type)
-                        if (arg_base, field_base) in CONVERSIONS:
-                            name = f"v{to_name(field_type)}"
-                            if isinstance(arg, ir.Var):
-                                name = arg.name + "_" + to_name(field_type)
-                            new_arg = f.var(name, field.type)
-                            conv = CONVERSIONS[(arg_base, field_base)]
-                            self.new_relations.add(conv)
-                            tasks.append(f.lookup(
-                                conv,
-                                [arg, new_arg]
-                            ))
+                        if arg_base is not None and field_base is not None and conversion_allowed(arg, arg_base, field_base):
+                            new_arg = self.convert(arg, arg_base, field_base, tasks)
                             final_args.append(new_arg)
                         else:
                             final_args.append(arg)
@@ -982,7 +1023,6 @@ class Replacer(visitor.Rewriter):
                         final_args.append(arg)
                 tasks.append(f.lookup(overload, final_args))
                 branches.append(f.logical(tasks, hoists))
-
         # unwrap if we don't actually need a union
         if len(branches) == 1:
             if len(branches[0].body) == 1:
@@ -992,6 +1032,48 @@ class Replacer(visitor.Rewriter):
         else:
             return f.union(branches, hoists)
 
+    def convert(self, arg:ir.Value, actual:ir.Type, expected:ir.Type, tasks:list[ir.Task]) -> ir.Value:
+        if type_matches(actual, expected):
+            return arg
+        value = None
+        if isinstance(arg, ir.Literal):
+            value = arg.value
+        elif isinstance(arg, (int, float, PyDecimal)):
+            value = arg
+        if value is not None:
+            if isinstance(value, int) and expected == types.Int:
+                return f.literal(value, expected)
+            if isinstance(value, (int, float, PyDecimal)) and expected == types.Float:
+                return float(value)
+            if isinstance(value, (int, float, PyDecimal)) and expected in (types.Decimal64, types.Decimal128):
+                # Converting str(value) rather than value avoids precision loss with Float->Decimal conversion.
+                return f.literal(PyDecimal(str(value)), expected)
+        name = f"v{to_name(expected)}"
+        if isinstance(arg, ir.Var):
+            name = arg.name + "_" + to_name(expected)
+        expected_base = to_base_primitive(expected) or expected
+        new_arg = f.var(name, expected_base)
+        tasks.append(f.lookup(builtins.cast, (expected_base, arg, new_arg)))
+        return new_arg
+
+    @staticmethod
+    def convert_literal_value(value, t: ir.Type):
+        if types.is_subtype(t, types.String):
+            return value, types.String
+        if isinstance(value, str) and types.is_subtype(t, types.DateTime):
+            return datetime.datetime.fromisoformat(value), types.DateTime
+        if isinstance(value, str) and types.is_subtype(t, types.Date):
+            return datetime.date.fromisoformat(value), types.Date
+        if isinstance(value, int) and types.is_subtype(t, types.Int):
+            return value, types.Int
+        if isinstance(value, (int, float, PyDecimal)) and types.is_subtype(t, types.Float):
+            return float(value), types.Float
+        if isinstance(value, (int, float, PyDecimal)) and types.is_subtype(t, types.Decimal64):
+            return PyDecimal(str(value)), types.Decimal64
+        if isinstance(value, (int, float, PyDecimal)) and types.is_subtype(t, types.Decimal128):
+            return PyDecimal(str(value)), types.Decimal128
+        return value, t
+
 #--------------------------------------------------
 # Typer pass
 #--------------------------------------------------
@@ -1021,6 +1103,6 @@ class InferTypes(compiler.Pass):
             final = Replacer(w.net).walk(model)
 
         for err in w.net.errors:
-            rich.print(str(err))
+            rich.print(str(err), file=sys.stderr)
 
         return final
