@@ -5,7 +5,9 @@ from decimal import Decimal as PyDecimal
 
 from relationalai.early_access.metamodel import ir, compiler as c, builtins as bt, types, visitor, helpers, factory as f
 from relationalai.early_access.metamodel.typer import Checker, InferTypes
+from relationalai.early_access.metamodel.typer.typer import to_base_primitive, to_type
 from relationalai.early_access.metamodel.rewrite import Flatten
+from relationalai.early_access.metamodel.rewrite import ExtractKeys
 
 from relationalai.early_access.metamodel.rewrite import Splinter
 from relationalai.early_access.metamodel.visitor import ReadWriteVisitor
@@ -13,6 +15,8 @@ from relationalai.early_access.metamodel.util import OrderedSet, group_by, NameC
 
 from relationalai.early_access.rel import rel, rel_utils as u, rewrite, builtins as rel_bt
 
+import math
+
 
 #--------------------------------------------------
 # Compiler
@@ -24,6 +28,7 @@ class Compiler(c.Compiler):
             Checker(),
             InferTypes(),
             rewrite.CDC(),
+            ExtractKeys(),
             # rewrite.ExtractCommon(),
             Flatten(),
             rewrite.QuantifyVars(),
@@ -57,87 +62,86 @@ class ModelToRel:
 
     def to_rel(self, model: ir.Model, options:dict = {}) -> rel.Program:
         self._register_external_relations(model)
-        if options.get("no_declares"):
-            return rel.Program(tuple(self._generate_rules(model)))
-        else:
-            return rel.Program(tuple([
-                *self._generate_builtin_defs(model),
-                *self._generate_declares(model),
-                *self._generate_rules(model),
-            ]))
+
+        rules = self._generate_rules(model)
+        reads = self._rel_reads(rules)
+        declares = [] if options.get("no_declares") else self._generate_declares(model)
+        self._rel_reads(declares, reads)
+        return rel.Program(tuple([
+            *self._generate_builtin_defs(model, reads),
+            *declares,
+            *rules,
+        ]))
 
     #--------------------------------------------------
     # Top level handlers
     #--------------------------------------------------
 
-    def _generate_builtin_defs(self, model) -> list[rel.Def]:
+    def _generate_builtin_defs(self, model, reads:OrderedSet[str]) -> list[rel.Def]:
         defs = []
 
-        defs.append(
-            rel.Def("pyrel_Decimal64",
-                tuple([rel.Var("x")]),
-                rel.atom("::std::common::FixedDecimal",
-                    tuple([rel.MetaValue(64), rel.MetaValue(u.DECIMAL64_SCALE), rel.Var("x")])),
-                tuple([rel.Annotation("inline", ())]),
-            ),
-        )
-
-        defs.append(
-            rel.Def("pyrel_Decimal128",
-                tuple([rel.Var("x")]),
-                rel.atom("::std::common::FixedDecimal",
-                    tuple([rel.MetaValue(128), rel.MetaValue(u.DECIMAL128_SCALE), rel.Var("x")])),
-                tuple([rel.Annotation("inline", ())]),
-            ),
-        )
-
-        defs.append(
-            rel.Def("pyrel_parse_decimal64",
-                tuple([rel.Var("x", type="::std::common::String"), rel.Var("y", type="pyrel_Decimal64")]),
-                rel.atom("::std::common::parse_decimal", tuple([rel.MetaValue(64), rel.MetaValue(u.DECIMAL64_SCALE), rel.Var("x"), rel.Var("y")])),
-                tuple([rel.Annotation("inline", ())]),
-            ),
-        )
-
-        defs.append(
-            rel.Def("pyrel_parse_decimal128",
-                tuple([rel.Var("x", type="::std::common::String"), rel.Var("y", type="pyrel_Decimal128")]),
-                rel.atom("::std::common::parse_decimal", tuple([rel.MetaValue(128), rel.MetaValue(u.DECIMAL128_SCALE), rel.Var("x"), rel.Var("y")])),
-                tuple([rel.Annotation("inline", ())]),
-            ),
-        )
-
-        for (r, from_type, to_type) in [
-            (bt.int_to_decimal64, "::std::common::Int", "pyrel_Decimal64"),
-            (bt.int_to_decimal128, "::std::common::Int", "pyrel_Decimal128"),
-            (bt.float_to_decimal64, "::std::common::Float", "pyrel_Decimal64"),
-            (bt.float_to_decimal128, "::std::common::Float", "pyrel_Decimal128"),
-            (bt.decimal64_to_float, "pyrel_Decimal64", "::std::common::Float"),
-            (bt.decimal128_to_float, "pyrel_Decimal128", "::std::common::Float"),
-            (bt.decimal64_to_decimal128, "pyrel_Decimal64", "pyrel_Decimal128"),
-            (bt.decimal128_to_decimal64, "pyrel_Decimal128", "pyrel_Decimal64"),
-        ]:
+        if "pyrel_Decimal64" in reads or "pyrel_parse_decimal64" in reads:
             defs.append(
-                self._convert_def(u.OPERATORS[r.name], from_type, to_type),
+                rel.Def("pyrel_Decimal64",
+                    tuple([rel.Var("x")]),
+                    rel.atom("::std::common::FixedDecimal",
+                        tuple([rel.MetaValue(64), rel.MetaValue(u.DECIMAL64_SCALE), rel.Var("x")])),
+                    tuple([rel.Annotation("inline", ())]),
+                ),
+            )
+
+        if "pyrel_Decimal128" in reads or "pyrel_parse_decimal128" in reads:
+            defs.append(
+                rel.Def("pyrel_Decimal128",
+                    tuple([rel.Var("x")]),
+                    rel.atom("::std::common::FixedDecimal",
+                        tuple([rel.MetaValue(128), rel.MetaValue(u.DECIMAL128_SCALE), rel.Var("x")])),
+                    tuple([rel.Annotation("inline", ())]),
+                ),
+            )
+
+        if "pyrel_parse_decimal64" in reads:
+            defs.append(
+                rel.Def("pyrel_parse_decimal64",
+                    tuple([rel.Var("x", type="::std::common::String"), rel.Var("y", type="pyrel_Decimal64")]),
+                    rel.atom("::std::common::parse_decimal", tuple([rel.MetaValue(64), rel.MetaValue(u.DECIMAL64_SCALE), rel.Var("x"), rel.Var("y")])),
+                    tuple([rel.Annotation("inline", ())]),
+                ),
+            )
+
+        if "pyrel_parse_decimal128" in reads:
+            defs.append(
+                rel.Def("pyrel_parse_decimal128",
+                    tuple([rel.Var("x", type="::std::common::String"), rel.Var("y", type="pyrel_Decimal128")]),
+                    rel.atom("::std::common::parse_decimal", tuple([rel.MetaValue(128), rel.MetaValue(u.DECIMAL128_SCALE), rel.Var("x"), rel.Var("y")])),
+                    tuple([rel.Annotation("inline", ())]),
+                ),
             )
 
         return defs
 
     @staticmethod
-    def _convert_def(name: str, input_type: str, output_type: str):
-        return rel.Def(name,
-            tuple([rel.Var("x", type=input_type), rel.Var("y", type=output_type)]),
-            rel.Exists(
-                tuple([rel.Var("type_x"), rel.Var("type_y")]),
-                rel.And(ordered_set(
-                    # Since we declared them to be the types we're converting from and to, we can just use the types of x and y here.
-                    # # The Rel compiler will use the static type of the variable to compute the Type values.
-                    rel.atom("rel_primitive_typeof", tuple([rel.Var("x"), rel.Var("type_x")])),
-                    rel.atom("rel_primitive_typeof", tuple([rel.Var("y"), rel.Var("type_y")])),
-                    rel.atom("rel_primitive_convert", tuple([rel.Var("type_x"), rel.Var("type_y"), rel.Var("x"), rel.Var("y")])),
-                )),
-            ),
-        )
+    def _convert_abs(from_type: ir.Type, to_type: ir.Type):
+        if from_type == types.Int and to_type == types.Float:
+            return rel.Identifier("::std::common::int_float_convert")
+        elif from_type == types.Float and to_type == types.Int:
+            return rel.Identifier("::std::common::float_int_convert")
+        else:
+            input_type = u.rel_typename(from_type)
+            output_type = u.rel_typename(to_type)
+            return rel.RelationalAbstraction(
+                tuple([rel.Var("x", type=input_type), rel.Var("y", type=output_type)]),
+                rel.Exists(
+                    tuple([rel.Var("type_x"), rel.Var("type_y")]),
+                    rel.And(ordered_set(
+                        # Since we declared them to be the types we're converting from and to, we can just use the types of x and y here.
+                        # The Rel compiler will use the static type of the variable to compute the Type values.
+                        rel.atom("rel_primitive_typeof", tuple([rel.Var("x"), rel.Var("type_x")])),
+                        rel.atom("rel_primitive_typeof", tuple([rel.Var("y"), rel.Var("type_y")])),
+                        rel.atom("rel_primitive_convert", tuple([rel.Var("type_x"), rel.Var("type_y"), rel.Var("x"), rel.Var("y")])),
+                    )),
+                )
+            )
 
     def _generate_declares(self, m: ir.Model) -> list[rel.Declare]:
         """
@@ -216,13 +220,16 @@ class ModelToRel:
                     return []
                 return [rel.RawSource(cast(str, effect.args[1]))]
             else:
-                args, lookups = self._effect_args(effect)
+                args, lookups, rel_equiv = self._effect_args(effect)
                 if lookups:
                     other.extend(lookups)
                 return [rel.Def(
                     self._effect_name(effect),
                     args,
-                    self.generate_logical_body(other, aggregates, ranks),
+                    rel.create_and([
+                        self.generate_logical_body(other, aggregates, ranks),
+                        *rel_equiv
+                    ]),
                     self.generate_annotations(effect.annotations)
                 )]
         else:
@@ -244,13 +251,16 @@ class ModelToRel:
             for name, updates in relation_groups.items():
                 effects_to_union = []
                 for update in updates:
-                    update_args, lookups = self._effect_args(update)
+                    update_args, lookups, rel_equiv = self._effect_args(update)
                     if update_args:
                         defs.append(
                             rel.Def(
                                 name,
                                 update_args,
-                                self.generate_body_expr(lookups),
+                                rel.create_and([
+                                    self.generate_body_expr(lookups),
+                                    *rel_equiv
+                                ]),
                                 self.generate_annotations(update.annotations)
                             )
                         )
@@ -259,7 +269,7 @@ class ModelToRel:
 
                 if effects_to_union:
                     update = updates.some()
-                    args, lookups = self._effect_args(update)
+                    args, lookups, rel_equiv = self._effect_args(update)
                     bodies = []
                     for update in effects_to_union:
                         bodies.append(self.handle(update))
@@ -270,7 +280,10 @@ class ModelToRel:
                         rel.Def(
                             name,
                             args,
-                            rel.Union(tuple(bodies)),
+                            rel.create_and([
+                                rel.Union(tuple(bodies)),
+                                *rel_equiv
+                            ]),
                             self.generate_annotations(update.annotations)
                         )
                     )
@@ -400,7 +413,8 @@ class ModelToRel:
     def handle(self, n: ir.Node):
         """ Dispatch to the appropriate ir.Node handler. """
         if isinstance(n, ir.PyValue):
-            return self.handle_value(type(n), n)
+            t = to_type(n)
+            return self.handle_value(t, n)
         handler = getattr(self, f"handle_{n.kind}", None)
         if handler:
             return handler(n)
@@ -411,30 +425,56 @@ class ModelToRel:
         """ Dispatch each node to the appropriate ir.Node handler. """
         return tuple([self.handle(x) for x in n])
 
-    def handle_value(self, type: Any, value: Any) -> Union[rel.Primitive, rel.RelationalAbstraction, rel.MetaValue, rel.Var]:
+    def handle_value(self, type: ir.Type|None, value: Any) -> Union[rel.Primitive, rel.RelationalAbstraction, rel.MetaValue, rel.Var]:
         """ Handle the value (Node or Value) and wrap in a Metavalue if the type is Symbol. """
         # only handle if it is a Node (e.g. ir.Var or ir.Literal)
         v = self.handle(value) if isinstance(value, ir.Node) else value
+
+        # type might be None for these so we have to handle them before the check below.
+        if isinstance(v, float) and (math.isinf(v) or math.isnan(v)):
+            x = rel.Var("_float")
+            rel_name = "::std::common::infinity" if math.isinf(v) else "::std::common::nan"
+            return rel.RelationalAbstraction(
+                tuple([x]),
+                rel.atom(rel_name, tuple([rel.MetaValue(64), x])),
+            )
+
+        if type is None:
+            return v
         # only wrap if v is a primitive (i.e. not a metavalue or a var, for example).
+        base = to_base_primitive(type) or type
         if type == types.Symbol and isinstance(v, (str, int, float, bool)):
             return rel.MetaValue(v)
-        elif type == types.Decimal64 and isinstance(v, PyDecimal):
-
-            # Use parse_decimal to avoid precision loss since Rel would interpret a decimal string as a float.
-            name = "_dec"
-            x = rel.Var(name)
-            return rel.RelationalAbstraction(
-                tuple([x]),
-                rel.atom("::std::common::parse_decimal", tuple([rel.MetaValue(64), rel.MetaValue(u.DECIMAL64_SCALE), str(v), x]))
-            )
-        elif type == types.Decimal128 and isinstance(v, PyDecimal):
-            # Use parse_decimal to avoid precision loss since Rel would interpret a decimal string as a float.
-            name = "_dec"
-            x = rel.Var(name)
-            return rel.RelationalAbstraction(
-                tuple([x]),
-                rel.atom("::std::common::parse_decimal", tuple([rel.MetaValue(128), rel.MetaValue(u.DECIMAL128_SCALE), str(v), x]))
-            )
+        elif base == types.Decimal64 and isinstance(v, PyDecimal):
+            if u.can_represent_as_int64(v):
+                # If v is a 64-bit integer, use the decimal constructor.
+                x = rel.Var("_dec")
+                return rel.RelationalAbstraction(
+                    tuple([x]),
+                    rel.atom("::std::common::decimal", tuple([rel.MetaValue(64), rel.MetaValue(u.DECIMAL64_SCALE), int(v), x]))
+                )
+            else:
+                # Use parse_decimal to avoid precision loss since Rel would interpret a decimal string as a float.
+                x = rel.Var("_dec")
+                return rel.RelationalAbstraction(
+                    tuple([x]),
+                    rel.atom("::std::common::parse_decimal", tuple([rel.MetaValue(64), rel.MetaValue(u.DECIMAL64_SCALE), str(v), x]))
+                )
+        elif base == types.Decimal128 and isinstance(v, PyDecimal):
+            if u.can_represent_as_int64(v):
+                # If v is really an int, use the decimal constructor.
+                x = rel.Var("_dec")
+                return rel.RelationalAbstraction(
+                    tuple([x]),
+                    rel.atom("::std::common::decimal", tuple([rel.MetaValue(128), rel.MetaValue(u.DECIMAL128_SCALE), int(v), x]))
+                )
+            else:
+                # Use parse_decimal to avoid precision loss since Rel would interpret a decimal string as a float.
+                x = rel.Var("_dec")
+                return rel.RelationalAbstraction(
+                    tuple([x]),
+                    rel.atom("::std::common::parse_decimal", tuple([rel.MetaValue(128), rel.MetaValue(u.DECIMAL128_SCALE), str(v), x]))
+                )
         else:
             return v
 
@@ -539,7 +579,18 @@ class ModelToRel:
             tuple(self.handle_value_list(self._relation_types(n.relation), n.args))
         )
 
+    def handle_cast(self, n: ir.Lookup):
+        assert len(n.args) == 3
+        (target_type, source, target) = n.args
+        assert isinstance(target_type, ir.Type), f"Expected Type, got {type(target_type)}"
+        from_type = to_type(source)
+        rel_abstraction = self._convert_abs(from_type, target_type)
+        types = (from_type, target_type)
+        return rel.Atom(rel_abstraction, tuple(self.handle_value_list(types, (source, target))))
+
     def handle_lookup(self, n: ir.Lookup):
+        if n.relation == bt.cast:
+            return self.handle_cast(n)
         # only translate names to Rel operators if the relation is a built-in
         name = self._relation_name(n.relation)
         if bt.is_builtin(n.relation):
@@ -606,7 +657,7 @@ class ModelToRel:
         else:
             raise Exception(f"Cannot retrieve effect name from node {type(n)}")
 
-    def _effect_args(self, n: ir.Task) -> Tuple[Tuple[Any], list[ir.Task]]:
+    def _effect_args(self, n: ir.Task) -> Tuple[Tuple[Any], list[ir.Task], list[rel.Expr]]:
         """
             Return the arguments for the head of an effect rule and a list of lookups to add
             to the body of the rule.
@@ -629,7 +680,7 @@ class ModelToRel:
         else:
             raise Exception(f"Cannot retrieve effect params from node {type(n)}")
 
-        args, lookups = [], []
+        args, lookups, rel_equiv = [], [], []
         for idx, handled in enumerate(handled_args):
             if handled is None:
                 var = ir.Var(types.Any, "head")
@@ -637,8 +688,9 @@ class ModelToRel:
                 lookups.append(f.lookup(bt.eq, [var, orig_args[idx]]))
             elif isinstance(handled, rel.RelationalAbstraction):
                 var = ir.Var(types.Any, "head")
-                args.append(self.handle(var))
-                lookups.append(f.lookup(bt.eq, [var, orig_args[idx]]))
+                rel_var = self.handle(var)
+                args.append(rel_var)
+                rel_equiv.append(rel.create_eq(rel_var, handled))
             elif not isinstance(handled, rel.Var):
                 args.append(handled)
             else:
@@ -649,6 +701,66 @@ class ModelToRel:
                     continue
                 # Deduplicate variable
                 new_var = ir.Var(types.Any, handled.name + "_dup" + str(cnt))
-                args.append(self.handle(new_var))
-                lookups.append(f.lookup(bt.eq, [new_var, orig_args[idx]]))
-        return tuple(args), lookups
+                rel_var = self.handle(new_var)
+                args.append(rel_var)
+                rel_equiv.append(rel.create_eq(rel_var, handled))
+        return tuple(args), lookups, rel_equiv
+
+    def _rel_reads(self, root, reads:OrderedSet[str]|None = None) -> OrderedSet[str]:
+        if reads is None:
+            reads = OrderedSet()
+
+        if isinstance(root, list):
+            for r in root:
+                self._rel_reads(r, reads)
+
+        elif isinstance(root, rel.Declare):
+            assert isinstance(root.premise, rel.Atom)
+            for arg in root.premise.args:
+                if (isinstance(arg, rel.Var) and
+                    arg.type is not None and
+                    not arg.type.startswith("::std::common::")):
+                    reads.add(arg.type)
+
+        elif isinstance(root, rel.Def):
+            self._rel_reads(root.body, reads)
+
+        elif isinstance(root, rel.Atom):
+            if isinstance(root.expr, rel.Identifier):
+                reads.add(root.expr.name)
+            self._rel_reads(root.args, reads)
+
+        elif isinstance(root, rel.RelationalAbstraction):
+            self._rel_reads(root.body, reads)
+
+        elif isinstance(root, rel.And):
+            for arg in root.body:
+                self._rel_reads(arg, reads)
+
+        elif isinstance(root, rel.Or):
+            for arg in root.body:
+                self._rel_reads(arg, reads)
+
+        elif isinstance(root, rel.Exists):
+            self._rel_reads(root.body, reads)
+
+        elif isinstance(root, rel.ForAll):
+            self._rel_reads(root.body, reads)
+
+        elif isinstance(root, rel.Not):
+            self._rel_reads(root.body, reads)
+
+        elif isinstance(root, rel.BinaryExpr):
+            self._rel_reads(root.lhs, reads)
+            self._rel_reads(root.rhs, reads)
+            reads.add(root.op)
+
+        elif isinstance(root, rel.Product):
+            for arg in root.body:
+                self._rel_reads(arg, reads)
+
+        elif isinstance(root, rel.Union):
+            for arg in root.body:
+                self._rel_reads(arg, reads)
+
+        return reads
