@@ -4,6 +4,7 @@ Helpers to generate Rel, many taken from the original PyRel implementation.
 from __future__ import annotations
 import re
 from relationalai.early_access.metamodel import ir, types, factory as f
+from decimal import Decimal as PyDecimal
 
 # Rel primitive relations used by the compiler
 # TODO: how to deal with varargs?
@@ -54,6 +55,10 @@ OPERATORS = {
     "natural_log": "::std::common::natural_log",
     "sqrt": "std::common::sqrt",
     "maximum": "::std::common::maximum",
+    "minimum": "::std::common::minimum",
+    "//": "::std::common::trunc_divide",
+    "isinf": "::std::common::Infinity",
+    "isnan": "::std::common::NaN",
 
     # strings
     "string": "::std::common::string",
@@ -71,20 +76,6 @@ OPERATORS = {
     "parse_decimal128": "pyrel_parse_decimal128",
     "decimal64": "rel_primitive_eq",
     "decimal128": "rel_primitive_eq",
-
-    # numeric conversions
-    "int_to_float": "::std::common::int_float_convert",
-    "int_to_decimal64": "pyrel_int_to_decimal64",
-    "int_to_decimal128": "pyrel_int_to_decimal128",
-    "float_to_int": "::std::common::float_int_convert",
-    "float_to_decimal64": "pyrel_float_to_decimal64",
-    "float_to_decimal128": "pyrel_float_to_decimal128",
-    "decimal64_to_int": "::std::common::decimal_int_convert",
-    "decimal64_to_float": "pyrel_decimal64_to_float",
-    "decimal64_to_decimal128": "pyrel_decimal64_to_decimal128",
-    "decimal128_to_int": "::std::common::decimal_int_convert",
-    "decimal128_to_float": "pyrel_decimal128_to_float",
-    "decimal128_to_decimal64": "pyrel_decimal128_to_decimal64",
 }
 
 def rel_operator(ir_op):
@@ -109,6 +100,8 @@ def rel_typename(ir_type: ir.Type):
             return '::std::common::UInt128'
         elif ir_type == types.RowId:
             return '::std::common::Int'
+        elif ir_type == types.Bool:
+            return '::std::common::Boolean'
         elif types.is_builtin(ir_type):
             return f'::std::common::{ir_type.name}'
         elif ir_type.super_types:
@@ -197,3 +190,28 @@ def sanitize_identifier(name: str) -> str:
     if  safe_name in RESERVED_WORDS:
         safe_name = safe_name + "_" # preferring the pythonic pattern of `from_` vs `_from`
     return safe_name
+
+def can_represent_as_int64(decimal_value: PyDecimal) -> bool:
+    """
+    Check if a Python Decimal can be represented exactly as a 64-bit signed integer.
+
+    Args:
+        decimal_value (PyDecimal): The input decimal value.
+
+    Returns:
+        bool: True if the decimal can be represented as a 64-bit signed integer, False otherwise.
+    """
+    # 64-bit signed integer range
+    INT64_MIN = -9223372036854775808  # -2^63
+    INT64_MAX = 9223372036854775807   # 2^63 - 1
+
+    # Check if the decimal has any fractional part
+    if decimal_value % 1 != 0:
+        return False
+
+    # Convert to integer and check range
+    try:
+        int_value = int(decimal_value)
+        return INT64_MIN <= int_value <= INT64_MAX
+    except (ValueError, OverflowError):
+        return False
