@@ -138,11 +138,13 @@ class SolverModel:
             assert sym_req, f"Cannot symbolify requirement {req} in `satisfy`"
             sym_reqs.append(sym_req)
         ser = b.union(*sym_reqs) if len(sym_reqs) > 1 else sym_reqs[0]
-        cons = self.Constraint.new(serialized=b.select(ser).where(*expr._where))
+        # TODO(coey) nested select not working properly on supply_chain, so have to put the where later, for now
+        # cons = self.Constraint.new(serialized=b.select(ser).where(*expr._where))
+        cons = self.Constraint.new(serialized=ser)
         defs = [cons]
         if name is not None:
             defs.append(cons.name(name))
-        b.define(*defs)
+        b.define(*defs).where(*expr._where)
         return cons
 
     # print counts of the number of model components
@@ -165,7 +167,10 @@ class SolverModel:
     def print(self, with_names: bool = False):
         print("Printing solver model.")
         # print variables
-        var_df = b.select(self.Variable.name | "_").to_df()
+        var_df = b.select(self.Variable.name | "_").where(self.Variable).to_df()
+        if var_df.empty:
+            print("No variables defined in the solver model.")
+            return None
         # TODO print type of variable also
         print(f"{var_df.shape[0]} variables:")
         print(var_df.to_string(index=False, header=False))
