@@ -1384,6 +1384,22 @@ class EngineResumeFailed(RAIException):
            f"You can also manually recreate the engine {suggest(suggestions)}"
         ])
 
+
+class InvalidEngineSizeError(RAIException):
+    def __init__(self, engine_size: str, valid_sizes: list[str]):
+        msg = f"Invalid engine size '{engine_size}'"
+        self.name = msg
+        self.message = msg
+
+        # Format the content and raw_content
+        self.content = "".join([
+            f"Invalid engine size [yellow]{engine_size}[/yellow] provided. Please check your config."
+            "\n",
+            f"Valid engine sizes are: [green]{', '.join(valid_sizes)}[/green]"
+        ])
+
+        super().__init__(self.message, self.name, self.content)
+
 class EngineNameValidationException(RAIException):
     def __init__(self, engine_name: str):
         self.engine_name = engine_name
@@ -1400,9 +1416,9 @@ class EngineNameValidationException(RAIException):
     def format_message(self):
         from relationalai.tools.cli_helpers import ENGINE_NAME_ERROR
         return "\n".join([
-            f"We tried to auto-create engine [yellow]{self.engine_name}[/yellow] but it failed. Engine name requirements are:"
+            f"We tried to auto-create engine [yellow]{self.engine_name}[/yellow] but it failed."
             "\n",
-            f"[yellow]{ENGINE_NAME_ERROR}"
+            f"Engine name requirements are: [yellow]{ENGINE_NAME_ERROR}[/yellow]"
         ])
 
 # --------------------------------------------------
@@ -1439,7 +1455,7 @@ def handle_missing_integration(e: Exception):
         isinstance(runtime_env, SnowbookEnvironment)
         and runtime_env.runner == "warehouse"
     )
-    if in_warehouse and "NameResolutionError" in str(e):
+    if in_warehouse and any((name in str(type(e)) for name in ("NameResolutionError", "ProgrammingError", "SnowflakeIntegrationMissingException"))):
         raise SnowflakeIntegrationMissingException() from None
 
 access_integration_warning = textwrap.dedent("""
@@ -1505,12 +1521,12 @@ class SnowflakeAppMissingException(RAIException):
             f"or your active role(s) don't have permission to access it.\n\n"
             f"If it's installed under a different name, run 'rai init' on the command line to set the app name."
         )
-        
+
         if self.role:
             base += f"\n\nYour current primary role is '{self.role}', and by default all assigned roles are used as secondary roles."
-        
+
         base += "\n\nWhile access via secondary roles is supported, we recommend granting the necessary privileges directly to your primary role to ensure consistent access."
-        
+
         return base
 
 
@@ -2003,6 +2019,43 @@ class SnowflakeProxySourceError(RAIException):
         Instead, use the `source` keyword argument when defining types. See https://relational.ai/docs/reference/python/Model/Type for details.
         """)
 
+#--------------------------------------------------
+# Direct Access Warnings
+#--------------------------------------------------
+
+class SessionResultFormatWarning(RAIWarning):
+    def __init__(self, use_direct_access: bool):
+        self.use_direct_access = use_direct_access
+        self.message = "Failed to alter the PYTHON_CONNECTOR_QUERY_RESULT_FORMAT session parameter."
+        self.name = "Session query result format mismatch"
+
+        # Format the content and raw_content
+        self.content = self.format_message()
+        self.raw_content = self.strip_rich_tags()
+
+        super().__init__(self.message, self.name, self.content)
+
+    def format_message(self):
+        if self.use_direct_access:
+            return textwrap.dedent(f"""
+            The PYTHON_CONNECTOR_QUERY_RESULT_FORMAT session parameter was not set to 'json'. Direct access will not work as expected.
+            Please ensure that the session is set to 'json' before using direct access.
+            If you provide the session make sure to construct with with the `json` result format:
+            [green] snowflake_session = Session.builder.configs({
+                "PYTHON_CONNECTOR_QUERY_RESULT_FORMAT": "json"
+                # other session parameters...
+            }).create() [/green]
+            """)
+        else:
+            return textwrap.dedent(f"""
+            The PYTHON_CONNECTOR_QUERY_RESULT_FORMAT session parameter was set to 'json' even though direct access is not used.
+            This might influence formatting of your results. To ensure default behavior you can construct your session like:
+            [green] snowflake_session = Session.builder.configs({
+                "PYTHON_CONNECTOR_QUERY_RESULT_FORMAT": "arrow"
+                # other session parameters...
+            }).create() [/green]
+            """)
+
 #--------------------------------------------------
 # Direct Access Exceptions
 #--------------------------------------------------
@@ -2046,3 +2099,27 @@ class InvalidAliasError(RAIException):
 
     def format_message(self):
         return self.message
+
+
+class LeftOverRelationException(RAIException):
+    def __init__(self, import_name: str, model: str):
+        self.import_name = import_name
+        self.model = model
+        self.message = f"Relations are not empty for import '{import_name}'"
+        self.name = f"Stream relations for '{import_name}' are not empty"
+        self.content = self.format_message()
+
+        super().__init__(self.message, self.name, self.content)
+
+    def format_message(self):
+        return textwrap.dedent(f"""
+        The import '{self.import_name}' cannot be created due to existing non-empty relationships.
+
+        You can use the [green]--force[/green] parameter to overwrite these:
+
+        [green]rai imports:stream --source {self.import_name.upper()} --model {self.model} --force[/green]
+
+        To delete the import and all its relationships, use the [green]--force[/green] parameter:
+
+        [green]rai imports:delete --object {self.import_name.upper()} --model {self.model} --force[/green]
+        """)
