@@ -6,6 +6,8 @@ from numbers import Number, Real
 import textwrap
 from typing import Any, Type, Union
 import warnings
+
+import numpy
 from ..errors import (
     DirectedGraphNotApplicable, DirectedGraphNotSupported,
     ParameterBoundBelowInclusive, ParameterBoundAboveInclusive,
@@ -128,6 +130,29 @@ class Compute:
         reachable = self._lib.reachable_from(self._graph, unwrap(node))
         return self._lookup(reachable)
 
+    # Class-level dictionary to track which models have had `full_reachable_from` installed.
+    # This is shared across all instances of Compute.
+    # Motivation: the class-level dict ensures that full_reachable_from gets installed once per model,
+    # not once per Compute.
+    _full_reachable_from_installed = {}
+
+    def full_reachable_from(self, node: dsl.Producer) -> dsl.Producer:
+        key = self._graph.model.name
+        if not self._full_reachable_from_installed.get(key, False):
+            query = """
+                namespace graphlib
+                    @track(:graphlib, :full_reachable_from)
+                    @outline
+                    def full_reachable_from({G}, u, v) : {
+                        G(:node, u) and (u = v or (exists((w) | full_reachable_from(G, u, w) and G(:edge, w, v))))
+                    }
+                end
+            """
+            self._graph.model.install_raw(query)
+            self._full_reachable_from_installed[key] = True
+        reachable = self._lib.full_reachable_from(self._graph, unwrap(node))
+        return self._lookup(reachable)
+
     def is_reachable(self, node1: dsl.Producer, node2: dsl.Producer) -> dsl.Expression:
         return self._lib.reachable_from(self._graph, unwrap(node1), unwrap(node2))
 
@@ -644,11 +669,14 @@ class Graph:
             if callable(v):
                 metadata[k] = v(metadata)
 
-        # JSON doesn't support decimals, so we need to convert them to
-        # floats before passing them along to the viz
+        # adjust some types as needed by JSON
         for k, v in metadata.items():
+            # decimals are not supported, convert to floats
             if isinstance(v, Decimal):
                 metadata[k] = float(v)
+            # numpy specific ints also not supported, convert to int
+            if isinstance(v, numpy.integer):
+                metadata[k] = int(v)
 
         # For some reason, the existance of "id" in the metadata
         # results in edges not getting displayed in the visualization,
@@ -683,7 +711,7 @@ class Graph:
                 "metadata": flat_style,
                 "nodes": {
                     node_id: {
-                        **({"label": info["label"]} if "label" in info else {}),
+                        **({"label": str(info["label"])} if "label" in info else {}),
                         "metadata": self._visual_props(style["node"], info.copy()),
                     }
                     for (node_id, info) in data["nodes"].items()
