@@ -217,7 +217,7 @@ def span_code(span:SpanNode):
         if toggles["rel"]:
             code_with_header("model rel", span.end_attrs["rel"], language="rel")
         if toggles["lqp"]:
-            code_with_header("lqp", span.end_attrs["lqp"], language="lqp")
+            code_with_header("lqp", span.end_attrs["lqp_debug"], language="lqp")
         if toggles["sql"]:
             code_with_header("sql", span.end_attrs["sql"], language="sql")
     elif span.type == "compile" and span.end_attrs.get("compile_type") == "query":
@@ -228,7 +228,7 @@ def span_code(span:SpanNode):
         if toggles["rel"]:
             code_with_header("rel", span.end_attrs["rel"], language="rel")
         if toggles["lqp"]:
-            code_with_header("lqp", span.end_attrs["lqp"], language="lqp")
+            code_with_header("lqp", span.end_attrs["lqp_debug"], language="lqp")
         if toggles["sql"]:
             code_with_header("sql", span.end_attrs["sql"], language="sql")
     elif span.type == "passes":
